"""Functions related to text generation of models."""

import logging
import sys
import warnings
from collections import defaultdict
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable

import torch
from datasets import Dataset
from torch.utils.data import DataLoader
from tqdm.auto import tqdm
from transformers import (
    GenerationConfig,
    PreTrainedTokenizerBase,
    StoppingCriteria,
    StoppingCriteriaList,
)
from transformers.modeling_utils import ModelOutput

from .exceptions import InvalidBenchmark
from .model_cache import (
    ModelCache,
    load_cached_model_outputs,
    split_dataset_into_cached_and_non_cached,
)
from .openai_models import OpenAIModel
from .structured_generation_utils import (
    get_ner_logits_processors,
    get_ner_prefix_allowed_tokens_fn,
)
from .tasks import NER
from .utils import SUPERTASKS_USING_LOGPROBS, clear_memory, get_end_of_chat_token_ids
from .vllm_models import VLLMModel

if TYPE_CHECKING:
    from transformers import DataCollator

    from .config import BenchmarkConfig, DatasetConfig, ModelConfig
    from .protocols import GenerativeModel, Tokenizer

logger = logging.getLogger(__package__)


def generate(
    itr: tqdm,
    prepared_train: Dataset,
    prepared_tests: list[Dataset],
    model: "GenerativeModel",
    model_config: "ModelConfig",
    tokenizer: "Tokenizer",
    data_collator: "DataCollator",
    compute_metrics: Callable,
    extract_labels_fn: Callable[..., list[Any]],
    benchmark_config: "BenchmarkConfig",
    dataset_config: "DatasetConfig",
) -> dict[str, list[dict[str, float]]]:
    """Evaluate a model on a dataset through generation.

    Args:
        itr:
            The progress bar iterator.
        prepared_train:
            The prepared training dataset.
        prepared_tests:
            The prepared bootstrapped test datasets.
        num_iter:
            The number of iterations to run.
        rng:
            The random number generator.
        model:
            The model to evaluate.
        model_config:
            The configuration of the model.
        tokenizer:
            The tokenizer to use for the model. If `None` then the model's
            tokenizer will be used.
        data_collator:
            The data collator to use for the model.
        compute_metrics:
            The function to use to compute the metrics.
        extract_labels_fn:
            The function to use to extract the labels from the model output.
        benchmark_config:
            The configuration of the benchmark.
        dataset_config:
            The configuration of the dataset.

    Returns:
        A dictionary containing the scores, with keys "test" and maybe "train", with
        values being lists of dicts containing the scores for each metric for each
        iteration.
    """
    scores: dict[str, list[dict[str, float]]] = defaultdict(list)

    # Set up the name of the model output cache. If we are testing then we save the
    # model outputs to a different cache and ensure that that cache is deleted before
    # the next test, to ensure that the tests are independent of each other
    model_cache_dir = Path(model_config.model_cache_dir)
    if not hasattr(sys, "_called_from_test"):
        cache_name = f"{dataset_config.name}-model-outputs.json"
    else:
        cache_name = f"{dataset_config.name}-model-outputs-test.json"
        (model_cache_dir / cache_name).unlink(missing_ok=True)

    cache = ModelCache(
        model_cache_dir=model_cache_dir,
        cache_name=cache_name,
        max_generated_tokens=dataset_config.max_generated_tokens,
    )

    for idx in itr:
        prepared_test = prepared_tests[idx]
        assert isinstance(prepared_test, Dataset)

        generation_kwargs = dict(
            model=model,
            tokenizer=tokenizer,
            data_collator=data_collator,
            compute_metrics=compute_metrics,
            extract_labels_fn=extract_labels_fn,
            dataset_config=dataset_config,
            cache=cache,
        )

        def update_scores(
            scores: dict[str, list[dict[str, float]]],
            benchmark_config: "BenchmarkConfig",
        ) -> dict[str, list[dict[str, float]]]:
            """Perform a single iteration of generation and update the scores.

            Args:
                scores:
                    The scores so far.
                benchmark_config:
                    The configuration of the benchmark.

            Returns:
                The updated scores.
            """
            test_scores = generate_single_iteration(
                prepared_dataset=prepared_test,
                benchmark_config=benchmark_config,
                **generation_kwargs,
            )
            logger.debug(f"Test scores for iteration {idx}: {test_scores}")
            scores["test"].append(test_scores)

            if benchmark_config.evaluate_train:
                train_scores = generate_single_iteration(
                    prepared_dataset=prepared_train,
                    benchmark_config=benchmark_config,
                    **generation_kwargs,
                )
                logger.debug(f"Train scores for iteration {idx}: {train_scores}")
                scores["train"].append(train_scores)

            clear_memory()
            return scores

        if isinstance(model, VLLMModel):
            scores = update_scores(scores=scores, benchmark_config=benchmark_config)
        else:
            while True:
                try:
                    scores = update_scores(
                        scores=scores, benchmark_config=benchmark_config
                    )
                    break
                except Exception as e:
                    oom_error = [
                        "CUDA out of memory",
                        "CUDA error",
                        "MPS backend out of memory",
                        "Too many parallel completions requested.",  # OpenAI specific
                    ]
                    if isinstance(model, VLLMModel) or all(
                        error not in str(e) for error in oom_error
                    ):
                        raise InvalidBenchmark(str(e))
                    clear_memory()
                    benchmark_config.batch_size //= 2
                    if benchmark_config.batch_size < 1:
                        raise InvalidBenchmark(
                            "GPU out of memory, even with a batch size of 1!"
                        )

    cache.remove()
    return scores


def generate_single_iteration(
    prepared_dataset: Dataset,
    model: "GenerativeModel",
    tokenizer: "Tokenizer",
    data_collator: "DataCollator",
    compute_metrics: Callable,
    extract_labels_fn: Callable[..., list[Any]],
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
    cache: ModelCache,
) -> dict[str, float]:
    """Evaluate a model on a dataset in a single iteration through generation.

    Args:
        prepared_dataset:
            The dataset to evaluate on.
        model:
            The model to evaluate.
        tokenizer:
            The tokenizer to use for the model.
        data_collator:
            The data collator to use for the model.
        compute_metrics:
            The function to use to compute the metrics.
        extract_labels_fn:
            The function to use to extract the labels from the dataset.
        dataset_config:
            The configuration of the dataset.
        benchmark_config:
            The configuration of the benchmark.
        cache:
            The model output cache.

    Returns:
        A list of dictionaries containing the scores for each metric.
    """
    cache.load()

    # Split up the prepared dataset into a cached and non-cached part
    cached_dataset, non_cached_dataset = split_dataset_into_cached_and_non_cached(
        dataset=prepared_dataset, cache=cache
    )

    all_preds: list[str | list[str]] = list()

    if len(non_cached_dataset) > 0:
        # Tokens used in generation to know when generation is finished
        stopping_criteria = get_generation_stopping_criteria(
            tokenizer=tokenizer, model=model
        )

        generation_config = GenerationConfig(
            # What to output
            max_new_tokens=dataset_config.max_generated_tokens,
            output_scores=dataset_config.task.supertask in SUPERTASKS_USING_LOGPROBS,
            return_dict_in_generate=True,
            # How to sample
            do_sample=False,  # Equivalent to greedy decoding (temperature=0)
            # Special tokens
            bos_token_id=tokenizer.bos_token_id,
            eos_token_id=tokenizer.eos_token_id,
            pad_token_id=tokenizer.pad_token_id,
        )

        # Sort the non_cached dataset by the length of the text, to minimise the amount
        # of padding that needs to be added, speeding up generation
        non_cached_dataset = non_cached_dataset.add_column(
            name="length", column=[len(x) for x in non_cached_dataset["text"]]
        )
        non_cached_dataset = non_cached_dataset.sort("length", reverse=True)

        # Enable batching by building a dataloader. The dataloader cannot deal with
        # text columns, so we create a copy of the dataset without these
        torch_dataset = non_cached_dataset.with_format("torch").remove_columns(
            [
                column
                for column in non_cached_dataset.column_names
                if column != "input_ids"
            ]
        )

        if isinstance(model, OpenAIModel):
            batch_size = 1
        elif isinstance(model, VLLMModel):
            batch_size = len(torch_dataset)
        else:
            batch_size = benchmark_config.batch_size

        dataloader = DataLoader(
            dataset=torch_dataset,
            batch_size=batch_size,
            shuffle=False,
            num_workers=4,
            collate_fn=data_collator,
        )

        with warnings.catch_warnings():
            # This ignores the following warning, which is out of our control:
            #   "os.fork() was called. os.fork() is incompatible with multithreaded
            #   code, and JAX is multithreaded, so this will likely lead to a deadlock."
            warnings.simplefilter("ignore", category=RuntimeWarning)

            itr = (
                dataloader
                if batch_size == len(torch_dataset)
                else tqdm(
                    iterable=dataloader,
                    leave=False,
                    disable=hasattr(sys, "_called_from_test"),
                )
            )

            # Generate the completions for the non-cached examples
            for batch_idx, batch in enumerate(itr):
                model_output, extracted_labels = generate_batch(
                    batch=batch,
                    batch_idx=batch_idx,
                    batch_size=batch_size,
                    non_cached_dataset=non_cached_dataset,
                    model=model,
                    tokenizer=tokenizer,
                    stopping_criteria=stopping_criteria,
                    generation_config=generation_config,
                    extract_labels_fn=extract_labels_fn,
                    dataset_config=dataset_config,
                )
                cache.add_to_cache(
                    model_input=batch["input_ids"],
                    model_output=model_output,
                    tokenizer=tokenizer,
                )
                all_preds.extend(extracted_labels)

        if isinstance(itr, tqdm):
            itr.close()

        # Store the cache to disk
        cache.save()

    # Fetch the cached predictions for the cached examples
    if len(cached_dataset) > 0:
        model_output = load_cached_model_outputs(
            cached_dataset=cached_dataset, cache=cache, tokenizer=tokenizer
        )
        extracted_labels = extract_labels_fn(
            input_batch=cached_dataset, model_output=model_output, tokenizer=tokenizer
        )
        all_preds.extend(extracted_labels)

    if "label" in non_cached_dataset.column_names:
        ground_truth = [
            label.lower() if isinstance(label, str) else label
            for label in non_cached_dataset["label"] + cached_dataset["label"]
        ]
    elif "labels" in non_cached_dataset.column_names:
        ground_truth = [
            [label.lower() if isinstance(label, str) else label for label in label_list]
            for label_list in non_cached_dataset["labels"] + cached_dataset["labels"]
        ]
    elif "target_text" in non_cached_dataset.column_names:
        ground_truth = non_cached_dataset["target_text"] + cached_dataset["target_text"]
    else:
        raise ValueError(
            "The dataset must have either a 'label', 'labels', or 'target_text' column"
        )

    itr_scores: dict[str, float] = compute_metrics(
        model_outputs_and_labels=(all_preds, ground_truth),
        id2label=dataset_config.id2label,
    )

    return itr_scores


class StopWordCriteria(StoppingCriteria):
    """Stopping criteria for generation based on stop words.

    Attributes:
        stop_word_id_lists:
            A list of lists of token IDs that are used to determine whether generation
            should stop.
        indices_done:
            A list of indices of the examples for which generation has already stopped.
            Resets every batch.
    """

    def __init__(self, stop_word_id_lists: list[list[int]]):
        """Initialize the stopping criteria.

        Args:
            stop_word_id_lists:
                A list of lists of token IDs that are used to determine whether
                generation should stop.
        """
        super().__init__()
        self.stop_word_id_lists = stop_word_id_lists
        self.indices_done: list[int] = list()

    def clear(self) -> None:
        """Clear the example indices for which generation has already stopped."""
        self.indices_done = list()

    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs
    ) -> bool:
        """Determine whether generation should stop.

        Args:
            input_ids:
                The input IDs of the generated sequences.
            scores:
                The scores of the generated sequences. Not used.
            **kwargs:
                Additional keyword arguments. Not used.

        Returns:
            Whether generation should stop.
        """
        for stop_word_id_list in self.stop_word_id_lists:
            for batch_idx in range(input_ids.shape[0]):
                inputs = input_ids[batch_idx].tolist()
                sample_ends_with_stop_word = (
                    inputs[-len(stop_word_id_list) :] == stop_word_id_list
                )
                if sample_ends_with_stop_word:
                    self.indices_done.append(batch_idx)
                if all(idx in self.indices_done for idx in range(input_ids.shape[0])):
                    return True
        return False


def generate_batch(
    batch: dict[str, torch.Tensor],
    batch_idx: int,
    batch_size: int,
    non_cached_dataset: Dataset,
    model: "GenerativeModel",
    tokenizer: "Tokenizer",
    stopping_criteria: StopWordCriteria,
    generation_config: GenerationConfig,
    extract_labels_fn: Callable[..., list[str]],
    dataset_config: "DatasetConfig",
) -> tuple[ModelOutput, list[str | list[str]]]:
    """Evaluate a model on a single batch of examples through generation.

    Args:
        batch:
            The batch of examples to evaluate on.
        batch_idx:
            The index of the batch.
        batch_size:
            The size of the batch.
        non_cached_dataset:
            The dataset to evaluate on.
        model:
            The model to evaluate.
        tokenizer:
            The tokenizer used to encode the examples.
        stopping_criteria:
            The stopping criteria to use to stop generation.
        generation_config:
            The generation configuration to use.
        extract_labels_fn:
            The function to use to extract the labels from the model output.
        dataset_config:
            The configuration of the dataset.

    Returns:
        The predictions generated so far, with the predictions for the current batch
        appended.
    """
    # Generate the completions of the documents in the batch
    with torch.inference_mode():
        inputs = batch["input_ids"].to(model.device)
        stopping_criteria.clear()

        prefix_allowed_tokens_fn = None
        logits_processors = None
        if dataset_config == NER and isinstance(tokenizer, PreTrainedTokenizerBase):
            ner_tag_names = list(dataset_config.prompt_label_mapping.values())
            prefix_allowed_tokens_fn = get_ner_prefix_allowed_tokens_fn(
                ner_tag_names=ner_tag_names, tokenizer=tokenizer
            )
            if isinstance(model, VLLMModel):
                logits_processors = get_ner_logits_processors(
                    ner_tag_names=ner_tag_names, llm=model
                )

        model_output = model.generate(
            inputs=inputs,
            generation_config=generation_config,
            stopping_criteria=StoppingCriteriaList([stopping_criteria]),
            prefix_allowed_tokens_fn=prefix_allowed_tokens_fn,
            logits_processors=logits_processors,
        )
        assert isinstance(model_output, ModelOutput)

    # Some models include the input in the generated sequence, so we need to remove the
    # input if it is present
    inputs = inputs.detach().cpu()
    model_output.sequences = model_output.sequences.detach().cpu()
    if torch.equal(model_output.sequences[:, : inputs.shape[1]], inputs):
        model_output.sequences = model_output.sequences[:, inputs.shape[1] :]

    # Extract the labels from the model output and store them for metric computation
    # later
    batch_start = batch_idx * batch_size
    batch_end = (batch_idx + 1) * batch_size
    input_batch = non_cached_dataset[batch_start:batch_end]
    extracted_labels: list = extract_labels_fn(
        input_batch=input_batch, model_output=model_output, tokenizer=tokenizer
    )

    return model_output, extracted_labels


def extract_raw_predictions(
    generated_sequences: torch.Tensor, tokenizer: "Tokenizer"
) -> list[str]:
    """Get the raw predictions from the generated sequences.

    Args:
        generated_sequences:
            The generated sequences from the model. The outer-most list is the
            batch dimension, the inner-most list is the sequence dimension,
            consisting of token IDs.
        tokenizer:
            The tokenizer used to generate the tokens.

    Returns:
        The candidate labels with the smallest edit distance to the predicted labels.
    """
    raw_predictions: list[str] = [
        tokenizer.decode(completion_ids.tolist(), skip_special_tokens=True).split(
            "\n\n"
        )[0]
        for completion_ids in generated_sequences.long()
    ]

    end_of_chat_token_ids = get_end_of_chat_token_ids(tokenizer=tokenizer)
    if end_of_chat_token_ids is not None:
        end_of_chat_token = tokenizer.decode(end_of_chat_token_ids).strip()
        raw_predictions = [
            raw_prediction.split(end_of_chat_token)[0]
            for raw_prediction in raw_predictions
        ]

    raw_predictions = [raw_prediction.strip() for raw_prediction in raw_predictions]

    return raw_predictions


def get_generation_stopping_criteria(
    tokenizer: "Tokenizer", model: "GenerativeModel"
) -> StopWordCriteria:
    """Get the stopping criteria for generation.

    Args:
        tokenizer:
            The tokenizer used to tokenize the stop words.
        model:
            The generative model, which we use to ensure the tensors are on the
            same device, and also determine whether stop words are needed, based on
            the model type.

    Returns:
        The stopping criteria for generation.
    """
    if isinstance(model, OpenAIModel):
        return StopWordCriteria(stop_word_id_lists=[])

    double_newline_ids: list[int] = tokenizer(
        text=["\n\n"], add_special_tokens=False
    ).input_ids[0]
    single_newline_ids: list[int] = tokenizer(
        text=["\n"], add_special_tokens=False
    ).input_ids[0]
    bos_token_ids: list[int] = tokenizer(
        text=[tokenizer.bos_token], add_special_tokens=False
    ).input_ids[0]
    eos_token_ids: list[int] = tokenizer(
        text=[tokenizer.eos_token], add_special_tokens=False
    ).input_ids[0]

    end_chat_token_ids = get_end_of_chat_token_ids(tokenizer=tokenizer)
    if end_chat_token_ids is None:
        end_chat_token_ids = []

    def remove_empty_tokens(token_id_list: list[int]) -> list[int]:
        return [
            token_id for token_id in token_id_list if tokenizer.decode([token_id]) != ""
        ]

    double_newline_ids = remove_empty_tokens(double_newline_ids)
    single_newline_ids = remove_empty_tokens(single_newline_ids)
    bos_token_ids = remove_empty_tokens(bos_token_ids)
    eos_token_ids = remove_empty_tokens(eos_token_ids)

    stop_word_id_lists = [
        double_newline_ids,
        single_newline_ids + single_newline_ids,
        bos_token_ids,
        eos_token_ids,
        end_chat_token_ids,
    ]

    return StopWordCriteria(stop_word_id_lists=stop_word_id_lists)



"""Callbacks for the Hugging Face Trainer."""

import sys
from collections.abc import Sized

from tqdm.auto import tqdm
from transformers.trainer_callback import ProgressCallback


class NeverLeaveProgressCallback(ProgressCallback):
    """Progress callback which never leaves the progress bar."""

    def __init__(self, *args, **kwargs):
        """Initialise the callback."""
        super().__init__(*args, **kwargs)

    def on_train_begin(self, args, state, control, **kwargs):
        """Callback actions when training begins."""
        if state.is_local_process_zero:
            desc = "Finetuning model"
            self.training_bar = tqdm(
                total=None,
                leave=False,
                desc=desc,
                disable=hasattr(sys, "_called_from_test"),
            )
        self.current_step = 0

    def on_step_end(self, args, state, control, **kwargs):
        """Callback actions when a training step ends."""
        if state.is_local_process_zero:
            self.training_bar.update(state.global_step - self.current_step)
            self.current_step = state.global_step

    def on_prediction_step(self, args, state, control, eval_dataloader=None, **kwargs):
        """Callback actions when a prediction step ends."""
        if eval_dataloader is None:
            return
        correct_dtype = isinstance(eval_dataloader.dataset, Sized)
        if state.is_local_process_zero and correct_dtype:
            if self.prediction_bar is None:
                desc = "Evaluating model"
                self.prediction_bar = tqdm(
                    total=len(eval_dataloader),
                    leave=False,
                    desc=desc,
                    disable=hasattr(sys, "_called_from_test"),
                )
            self.prediction_bar.update(1)



"""Factory which produces datasets from a configuration."""

from typing import TYPE_CHECKING, Type

from .dataset_configs import get_dataset_config
from .utils import get_class_by_name

if TYPE_CHECKING:
    from .benchmark_dataset import BenchmarkDataset
    from .config import BenchmarkConfig, DatasetConfig


class DatasetFactory:
    """Factory which produces datasets from a configuration.

    Attributes:
        benchmark_config:
            The benchmark configuration to be used in all datasets constructed.
    """

    def __init__(self, benchmark_config: "BenchmarkConfig") -> None:
        """Initialize the dataset factory.

        Args:
            benchmark_config:
                The benchmark configuration to be used in all datasets constructed.
        """
        self.benchmark_config = benchmark_config

    def build_dataset(self, dataset: "str | DatasetConfig") -> "BenchmarkDataset":
        """Build a dataset from a configuration or a name.

        Args:
            dataset:
                The name of the dataset, or the dataset configuration.

        Returns:
            BenchmarkDataset:
                The benchmark dataset.
        """
        # Get the dataset configuration
        if isinstance(dataset, str):
            dataset_config = get_dataset_config(dataset)
        else:
            dataset_config = dataset

        # Get the benchmark class based on the task
        potential_class_names = [
            dataset_config.name,
            dataset_config.task.name,
            dataset_config.task.supertask,
        ]
        benchmark_cls: Type["BenchmarkDataset"] | None = get_class_by_name(
            class_name=potential_class_names
        )
        if not benchmark_cls:
            raise ValueError(
                "Could not find a benchmark class for any of the following potential "
                f"names: {', '.join(potential_class_names)}."
            )

        # Create the dataset
        dataset_obj = benchmark_cls(
            dataset_config=dataset_config, benchmark_config=self.benchmark_config
        )

        return dataset_obj



"""Functions related to the finetuning of models."""

import importlib.util
import logging
import sys
from collections import defaultdict
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, Type

import torch
from datasets import Dataset
from tqdm.auto import tqdm
from transformers import (
    EarlyStoppingCallback,
    IntervalStrategy,
    PreTrainedModel,
    PrinterCallback,
    ProgressCallback,
    TrainingArguments,
)
from transformers.trainer import OptimizerNames

from .callbacks import NeverLeaveProgressCallback
from .enums import DataType
from .exceptions import InvalidBenchmark, NaNValueInModelOutput
from .model_loading import load_model
from .utils import block_terminal_output, clear_memory, enforce_reproducibility

if TYPE_CHECKING:
    from transformers import DataCollator, Trainer

    from .config import BenchmarkConfig, DatasetConfig, ModelConfig
    from .protocols import Tokenizer

logger = logging.getLogger(__package__)


def finetune(
    itr: tqdm,
    train: Dataset,
    val: Dataset,
    tests: list[Dataset],
    prepared_train: Dataset,
    prepared_val: Dataset,
    prepared_tests: list[Dataset],
    model: PreTrainedModel,
    tokenizer: "Tokenizer",
    model_config: "ModelConfig",
    benchmark_config: "BenchmarkConfig",
    dataset_config: "DatasetConfig",
    compute_metrics: Callable,
    data_collator: "DataCollator",
    trainer_class: Type["Trainer"],
    evaluate_inputs_fn: Callable[..., dict[str, Any]],
    preprocess_logits_for_metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> dict[str, list[dict[str, float]]]:
    """Evaluate a model on a dataset through finetuning.

    Args:
        itr:
            The progress bar iterator.
        train:
            The training dataset.
        val:
            The validation dataset.
        tests:
            The bootstrapped test datasets.
        prepared_train:
            The prepared training dataset.
        prepared_val:
            The prepared validation dataset.
        prepared_tests:
            The prepared bootstrapped test datasets.
        model:
            The model to evaluate.
        tokenizer:
            The tokenizer to use.
        model_config:
            The configuration of the model.
        benchmark_config:
            The benchmark configuration.
        dataset_config:
            The dataset configuration.
        compute_metrics:
            The function used to compute the metrics.
        data_collator:
            The data collator to use.
        trainer_class:
            The Trainer class to use.
        evaluate_inputs_fn:
            A function that generates the appropriate inputs for the `Trainer.evaluate`
            method.
        preprocess_logits_for_metrics:
            A function that preprocesses the logits before they are passed to the
            `compute_metrics` function. This helps prevent memory issues during
            evaluation.

    Returns:
        A dictionary containing the scores, with keys "test" and maybe "train", with
        values being lists of dicts containing the scores for each metric for each
        iteration.
    """
    scores: dict[str, list[dict[str, float]]] = defaultdict(list)

    using_cuda = benchmark_config.device == torch.device("cuda")
    if using_cuda and torch.cuda.is_bf16_supported():
        dtype = DataType.BF16
    elif using_cuda:
        dtype = DataType.FP16
    else:
        dtype = DataType.FP32

    bs: int = benchmark_config.batch_size
    for idx in itr:
        # Set variable that tracks whether we need to initialize new models in
        # the single iteration call
        model_already_initialized = idx == 0

        # Run a loop here to deal with automatic reduction of batch size
        while True:
            # Clear GPU memory
            if not model_already_initialized:
                try:
                    del model
                except UnboundLocalError:
                    pass
                try:
                    del tokenizer
                except UnboundLocalError:
                    pass
                clear_memory()

            try:
                test = tests[idx]
                prepared_test = prepared_tests[idx]
                assert isinstance(test, Dataset)
                assert isinstance(prepared_test, Dataset)

                # Re-block terminal output, as it gets unblocked by the `transformers`
                # package before training
                block_terminal_output()

                training_args = get_training_args(
                    benchmark_config=benchmark_config,
                    model_config=model_config,
                    iteration_idx=idx,
                    dtype=dtype,
                    batch_size=bs,
                )

                itr_scores = finetune_single_iteration(
                    iteration_idx=idx,
                    model_config=model_config,
                    train=train,
                    prepared_train=prepared_train,
                    prepared_val=prepared_val,
                    test=test,
                    prepared_test=prepared_test,
                    training_args=training_args,
                    benchmark_config=benchmark_config,
                    dataset_config=dataset_config,
                    data_collator=data_collator,
                    compute_metrics=compute_metrics,
                    tokenizer=tokenizer if model_already_initialized else None,
                    model=model if model_already_initialized else None,
                    trainer_class=trainer_class,
                    evaluate_inputs_fn=evaluate_inputs_fn,
                    preprocess_logits_for_metrics=preprocess_logits_for_metrics,
                )

                if "train" in itr_scores:
                    logger.debug(
                        f"Train scores for iteration {idx}: {itr_scores['train']}"
                    )
                    scores["train"].append(itr_scores["train"])
                scores["test"].append(itr_scores["test"])
                logger.debug(f"Test scores for iteration {idx}: {itr_scores['test']}")

                break

            # NaN values can appear in the model output when using mixed precision, as
            # the hidden states get overflowed. In this case we try to disable mixed
            # precision and try again.
            except NaNValueInModelOutput:
                if dtype != DataType.FP32:
                    dtype = DataType.FP32
                    model_already_initialized = False
                    logger.debug(
                        "NaN value detected in model outputs while using mixed "
                        "precision. Retrying with full fp32 precision."
                    )
                else:
                    raise InvalidBenchmark(
                        "NaN value detected in model outputs, even with mixed "
                        "precision disabled."
                    )

            except Exception as e:
                if "CUDA" not in str(e) and "out of memory" not in str(e):
                    raise InvalidBenchmark(str(e))

                if bs <= 1:
                    msg = "Could not benchmark the model, even with a batch size of 1!"
                    if "MPS" in str(e):
                        msg += (
                            " As you are using MPS, you can try running the evaluation "
                            "with the `PYTORCH_MPS_HIGH_WATERMARK_RATIO=0.0` "
                            "environment variable set, as this removes the upper bound "
                            "on the memory usage."
                        )
                    raise InvalidBenchmark(msg)

                model_already_initialized = False

                bs //= 2
                logger.debug(f"Reduced batch size to {bs}")

    return scores


def finetune_single_iteration(
    iteration_idx: int,
    model_config: "ModelConfig",
    train: Dataset,
    test: Dataset,
    prepared_train: Dataset,
    prepared_val: Dataset,
    prepared_test: Dataset,
    training_args: TrainingArguments,
    benchmark_config: "BenchmarkConfig",
    dataset_config: "DatasetConfig",
    data_collator: "DataCollator",
    compute_metrics: Callable,
    tokenizer: "Tokenizer | None",
    model: PreTrainedModel | None,
    trainer_class: Type["Trainer"],
    evaluate_inputs_fn: Callable[..., dict[str, Any]],
    preprocess_logits_for_metrics: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
) -> dict[str, dict[str, float]]:
    """Run a single iteration of a benchmark.

    Args:
        iteration_idx:
            The index of the iteration.
        model_config:
            The model configuration.
        train:
            The original training dataset.
        test:
            The original test dataset.
        prepared_train:
            The prepared training dataset.
        prepared_val:
            The prepared validation dataset.
        prepared_test:
            The prepared test dataset.
        training_args:
            The training arguments.
        benchmark_config:
            The benchmark configuration.
        dataset_config:
            The dataset configuration.
        data_collator:
            The data collator.
        compute_metrics:
            The function to compute the metrics.
        tokenizer:
            The tokenizer to use in the benchmark. If None then a new tokenizer
            will be loaded.
        model:
            The model to use in the benchmark. If None then a new model will be
            loaded.
        trainer_class:
            The trainer class to use.
        evaluate_inputs_fn:
            A function that generates the appropriate inputs for the `Trainer.evaluate`
            method.
        preprocess_logits_for_metrics:
            A function that preprocesses the logits before they are passed to the
            `compute_metrics` function. This helps prevent memory issues during
            evaluation.

    Returns:
        A dictionary containing the scores for the current iteration, with keys `train`
        and `test`.
    """
    scores: dict[str, dict[str, float]] = dict()

    # Set random seeds to enforce reproducibility of the randomly initialised
    # weights
    seed = 4242 + iteration_idx
    enforce_reproducibility(framework=model_config.framework, seed=seed)

    if tokenizer is None or model is None:
        model_or_generative_model, tokenizer = load_model(
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )
        assert isinstance(model_or_generative_model, PreTrainedModel)
        model = model_or_generative_model

    compute_metrics = partial(compute_metrics, id2label=dataset_config.id2label)
    early_stopping = EarlyStoppingCallback(early_stopping_patience=2)
    trainer = trainer_class(
        model=model,
        args=training_args,
        train_dataset=prepared_train,
        eval_dataset=prepared_val,
        tokenizer=tokenizer,
        compute_metrics=compute_metrics,
        callbacks=[early_stopping],
        data_collator=data_collator,
        preprocess_logits_for_metrics=preprocess_logits_for_metrics,
    )

    if not benchmark_config.verbose:

        def no_logging(logs: dict[str, float]) -> None:
            return

        trainer.log = no_logging

    # Re-block terminal output, as it gets unblocked by the `transformers`
    # package before training
    block_terminal_output()

    # Sort out callbacks. We remove the callbacks that are producing unnecessary
    # output, to avoid cluttering the terminal output
    if not benchmark_config.verbose:
        trainer.remove_callback(PrinterCallback)
    trainer.remove_callback(ProgressCallback)
    if benchmark_config.progress_bar:
        trainer.add_callback(NeverLeaveProgressCallback)

    try:
        trainer.train()

        if benchmark_config.evaluate_train:
            with torch.inference_mode():
                evaluate_inputs = evaluate_inputs_fn(
                    dataset=train,
                    prepared_dataset=prepared_train,
                    metric_key_prefix="train",
                )
                train_scores = trainer.evaluate(**evaluate_inputs)
            scores["train"] = train_scores

        with torch.inference_mode():
            evaluate_inputs = evaluate_inputs_fn(
                dataset=test, prepared_dataset=prepared_test, metric_key_prefix="test"
            )
            test_scores = trainer.evaluate(**evaluate_inputs)
        scores["test"] = test_scores

        return scores

    except NaNValueInModelOutput as e:
        del trainer
        del model
        del tokenizer
        clear_memory()
        raise e

    except (RuntimeError, ValueError, IndexError) as e:
        raise InvalidBenchmark(str(e))


def get_training_args(
    benchmark_config: "BenchmarkConfig",
    model_config: "ModelConfig",
    iteration_idx: int,
    dtype: DataType,
    batch_size: int | None = None,
) -> TrainingArguments:
    """Get the training arguments for the current iteration.

    Args:
        benchmark_config:
            The benchmark configuration.
        model_config:
            The model configuration.
        iteration_idx:
            The index of the current iteration. This is only used to generate a
            unique random seed for the current iteration.
        dtype:
            The data type to use for the model weights.
        batch_size:
            The batch size to use for the current iteration, or None if the batch size
            in the benchmark config should be used.

    Returns:
        The training arguments for the current iteration.
    """
    # Set the logging strategy
    if benchmark_config.verbose:
        logging_strategy = IntervalStrategy.STEPS
    else:
        logging_strategy = IntervalStrategy.NO

    # Set seed variable
    seed = 4242 + iteration_idx

    if batch_size is None:
        batch_size = benchmark_config.batch_size

    if (
        benchmark_config.device == torch.device("cuda")
        and importlib.util.find_spec("bitsandbytes") is not None
    ):
        optimizer = OptimizerNames.ADAMW_8BIT
    else:
        optimizer = OptimizerNames.ADAMW_TORCH

    training_args = TrainingArguments(
        output_dir=model_config.model_cache_dir,
        evaluation_strategy=IntervalStrategy.STEPS,
        logging_strategy=logging_strategy,
        save_strategy=IntervalStrategy.STEPS,
        eval_steps=30,
        logging_steps=30,
        save_steps=30,
        max_steps=1 if hasattr(sys, "_called_from_test") else 10_000,
        use_cpu=benchmark_config.device == torch.device("cpu"),
        report_to=[],
        save_total_limit=1,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        learning_rate=2e-5,
        warmup_ratio=0.01,
        gradient_accumulation_steps=32 // batch_size,
        load_best_model_at_end=True,
        optim=optimizer,
        seed=seed,
        fp16=dtype == DataType.FP16,
        bf16=dtype == DataType.BF16,
        disable_tqdm=not benchmark_config.progress_bar,
        ddp_find_unused_parameters=False,
    )

    # TEMP: Use only 1 GPU for now for finetuning
    if benchmark_config.device == torch.device("cuda"):
        training_args._n_gpu = 1

    return training_args



"""Exceptions to used by other functions."""


class InvalidBenchmark(Exception):
    """The (model, dataset) combination cannot be benchmarked."""

    def __init__(
        self, message: str = "This model cannot be benchmarked on the given dataset."
    ):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class InvalidModel(Exception):
    """The model cannot be benchmarked on any datasets."""

    def __init__(
        self, message: str = "The model cannot be benchmarked on any datasets."
    ):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class HuggingFaceHubDown(Exception):
    """The Hugging Face Hub seems to be down."""

    def __init__(self, message: str = "The Hugging Face Hub is currently down."):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class NoInternetConnection(Exception):
    """There seems to be no internet connection."""

    def __init__(self, message: str = "There is currently no internet connection."):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class NaNValueInModelOutput(Exception):
    """There is a NaN value in the model output."""

    def __init__(self, message: str = "There is a NaN value in the model output."):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class FlashAttentionNotInstalled(Exception):
    """The `flash-attn` package has not been installed."""

    def __init__(
        self,
        message: str = (
            "The model you are trying to load requires Flash Attention. To use Flash "
            "Attention, please install the `flash-attn` package, which can be done by "
            "running `pip install -U wheel && FLASH_ATTENTION_SKIP_CUDA_BUILD=TRUE "
            "pip install flash-attn --no-build-isolation`."
        ),
    ):
        """Initialize the exception.

        Args:
            message:
                The message to display.
        """
        self.message = message
        super().__init__(self.message)


class NeedsExtraInstalled(InvalidModel):
    """The evaluation requires extra to be installed."""

    def __init__(self, extra: str):
        """Initialize the exception.

        Args:
            extra:
                The extra that needs to be installed.
        """
        self.message = (
            f"The model you are trying to load requires the `{extra}` extra to be "
            f"installed. To install the `{extra}` extra, please run `pip install "
            f"scandeval[{extra}]` or `pip install scandeval[all]`."
        )
        super().__init__(self.message)


class NeedsAdditionalArgument(InvalidModel):
    """The evaluation requires additional arguments to the `scandeval` command."""

    def __init__(self, cli_argument: str, script_argument: str, run_with_cli: bool):
        """Initialize the exception.

        Args:
            cli_argument:
                The argument that needs to be passed to the `scandeval` command.
            script_argument:
                The argument that needs to be passed to the `Benchmarker` class.
            run_with_cli:
                Whether the benchmark is being run with the CLI.
        """
        if run_with_cli:
            self.message = (
                f"The model you are trying to load requires the `{cli_argument}` "
                "argument to be passed to the `scandeval` command. Please pass the "
                "argument and try again."
            )
        else:
            self.message = (
                f"The model you are trying to load requires the `{script_argument}` "
                "argument  to be passed to the `Benchmarker` class. Please pass the "
                "argument and try again."
            )
        super().__init__(self.message)


class MissingHuggingFaceToken(InvalidModel):
    """The evaluation requires a Hugging Face token."""

    def __init__(self, run_with_cli: bool):
        """Initialize the exception.

        Args:
            run_with_cli:
                Whether the benchmark is being run with the CLI.
        """
        self.message = (
            "The model you are trying to load requires a Hugging Face token. "
        )
        if run_with_cli:
            self.message += (
                "Please run `huggingface-cli login` to login to the Hugging Face "
                "Hub and try again. Alternatively, you can pass your Hugging Face Hub "
                "token directly to the `scandeval` command with the `--token <token>` "
                "argument."
            )
        else:
            self.message += (
                "Please pass your Hugging Face Hub token to the `Benchmarker` class "
                "with the `token=<token>` argument  to the `Benchmarker` class and try "
                "again. Alternatively, you can also simply pass `token=True` (which is "
                "the default) to the `Benchmarker` class, which assumes that you have "
                "logged into the Hugging Face Hub. This can be done by running "
                "`huggingface-cli login` in the terminal or `from huggingface_hub "
                "import login; login()` in a Python script."
            )
        super().__init__(self.message)



"""Utility functions related to structured generation."""

import importlib.util
from typing import TYPE_CHECKING, Any, Callable

import torch
from pydantic import BaseModel, conlist, create_model

from .exceptions import InvalidBenchmark

if importlib.util.find_spec("outlines") is not None:
    # Check outlines version and make sure that it is at least 0.0.37
    from outlines._version import __version__ as outlines_version

    outlines_version = tuple(map(int, outlines_version.split(".")))
    if outlines_version < (0, 0, 37):
        raise InvalidBenchmark(
            "Evaluating models on the named entity recognition task requires "
            "`outlines>=0.0.37`. Please upgrade this with `pip install -U outlines` "
            "and try again. This is a temporary measure and will be installed properly "
            "in the next release."
        )

    try:
        from outlines.integrations.transformers import JSONPrefixAllowedTokens
        from outlines.integrations.vllm import JSONLogitsProcessor

    # TEMP: Remove this once `vllm` has a more lax dependency on `outlines`
    except ImportError:
        from outlines.serve.vllm import JSONLogitsProcessor

        JSONPrefixAllowedTokens = None

if TYPE_CHECKING:
    from transformers import PreTrainedTokenizerBase
    from vllm import LLM


def get_ner_schema(ner_tag_names: list[str]) -> type[BaseModel]:
    """Get the schema for the NER answer format, used for structured generation.

    Args:
        ner_tag_names:
            The NER tag names.

    Returns:
        The schema for the NER answer format.
    """
    keys_and_their_types: dict[str, Any] = {
        tag_name: (conlist(str, max_length=5), ...) for tag_name in ner_tag_names
    }
    schema = create_model("AnswerFormat", **keys_and_their_types)
    return schema


def get_ner_prefix_allowed_tokens_fn(
    ner_tag_names: list[str], tokenizer: "PreTrainedTokenizerBase"
) -> Callable[[int, torch.Tensor], list[int]]:
    """Get the prefix allowed tokens function for the NER task, used in `transformers`.

    Args:
        ner_tag_names:
            The NER tag names.
        tokenizer:
            The tokenizer to use for tokenizing the JSON Schema.

    Returns:
        The prefix allowed tokens function for the NER task.
    """
    # TEMP: Remove this once `vllm` has a more lax dependency on `outlines`
    if JSONPrefixAllowedTokens is None:
        raise InvalidBenchmark(
            "Evaluating models on the named entity recognition task without vLLM "
            "requires `outlines>=0.0.35`. Please upgrade this with `pip install -U "
            "outlines` and try again."
        )

    return JSONPrefixAllowedTokens(
        schema=get_ner_schema(ner_tag_names=ner_tag_names),
        tokenizer_or_pipe=tokenizer,
        whitespace_pattern=r" ?",
    )


def get_ner_logits_processors(
    ner_tag_names: list[str], llm: "LLM"
) -> list[Callable[[list[int], torch.Tensor], torch.Tensor]]:
    """Get the logits processors for the NER task, used in vLLM.

    Args:
        ner_tag_names:
            The NER tag names.
        llm:
            The vLLM model.

    Returns:
        The logit processors for the NER task.
    """
    logits_processor = JSONLogitsProcessor(
        schema=get_ner_schema(ner_tag_names=ner_tag_names),
        llm=llm,
        whitespace_pattern=r" ?",
    )
    return [logits_processor]



"""The different types of models that can be benchmarked."""

from typing import TYPE_CHECKING, Type

from .fresh import FreshModelSetup
from .hf import HFModelSetup
from .local import LocalModelSetup
from .openai import OpenAIModelSetup

if TYPE_CHECKING:
    from ..protocols import ModelSetup


# Note that the order of the model setup classes is important, as the first model setup
# that can load a model will be used
MODEL_SETUP_CLASSES: list[Type["ModelSetup"]] = [
    FreshModelSetup,
    LocalModelSetup,
    HFModelSetup,
    OpenAIModelSetup,
]



"""All dataset configurations used in ScandEval."""

from .config import DatasetConfig
from .languages import DA, DE, EN, FO, IS, NB, NL, NN, NO, SV, get_all_languages
from .tasks import COMMON_SENSE, KNOW, LA, NER, QA, SENT, SPEED, SUMM


def get_all_dataset_configs() -> dict[str, DatasetConfig]:
    """Get a mapping of all the dataset configurations.

    Returns:
        A mapping between names of datasets and their configurations.
    """
    dataset_configs = [
        cfg for cfg in globals().values() if isinstance(cfg, DatasetConfig)
    ]
    assert len(dataset_configs) == len({cfg.name for cfg in dataset_configs}), (
        "There are duplicate dataset configurations. Please ensure that each dataset "
        "has a unique name."
    )
    return {cfg.name: cfg for cfg in dataset_configs}


def get_dataset_config(dataset_name: str) -> DatasetConfig:
    """Get the dataset configuration for a dataset.

    Args:
        dataset_name:
            The name of the dataset.

    Returns:
        The dataset configuration.

    Raises:
        ValueError:
            If the dataset is not found.
    """
    # Get mapping of all dataset configs
    dataset_configs = get_all_dataset_configs()

    # If there are no matches for the dataset name, raise an error
    if dataset_name not in dataset_configs:
        raise ValueError(f"No dataset config found for dataset {dataset_name}.")

    # Otherwise, return the dataset configuration
    return dataset_configs[dataset_name]


### SENTIMENT DATASETS ###

SWEREC_CONFIG = DatasetConfig(
    name="swerec",
    pretty_name="the truncated version of the Swedish sentiment classification "
    "dataset SweReC",
    huggingface_id="ScandEval/swerec-mini",
    task=SENT,
    languages=[SV],
    prompt_prefix="Följande är recensioner och deras sentiment, som kan vara "
    "'positiv', 'neutral' eller 'negativ'.",
    prompt_template="Recension: {text}\nSentiment: {label}",
    prompt_label_mapping=dict(
        positive="positiv", neutral="neutral", negative="negativ"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

ANGRY_TWEETS_CONFIG = DatasetConfig(
    name="angry-tweets",
    pretty_name="the truncated version of the Danish sentiment classification "
    "dataset AngryTweets",
    huggingface_id="ScandEval/angry-tweets-mini",
    task=SENT,
    languages=[DA],
    prompt_prefix="Følgende er tweets og deres sentiment, som kan være 'positiv', "
    "'neutral' eller 'negativ'.",
    prompt_template="Tweet: {text}\nSentiment: {label}",
    prompt_label_mapping=dict(
        positive="positiv", neutral="neutral", negative="negativ"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

NOREC_CONFIG = DatasetConfig(
    name="norec",
    pretty_name="the truncated version of the Norwegian sentiment classification "
    "dataset NoReC",
    huggingface_id="ScandEval/norec-mini",
    task=SENT,
    languages=[NB, NN, NO],
    prompt_prefix="Følgende er anmeldelser og deres sentiment, som kan være 'positiv', "
    "'nøytral' eller 'negativ'.",
    prompt_template="Anmeldelse: {text}\nSentiment: {label}",
    prompt_label_mapping=dict(
        positive="positiv", neutral="nøytral", negative="negativ"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

# ISREC_CONFIG = DatasetConfig(
#     name="isrec",
#     pretty_name="the truncated version of IsReC",
#     huggingface_id="ScandEval/isrec-mini",  # TODO: Needs to be uploaded
#     task=SENT,
#     languages=[IS],
#     prompt_prefix="Eftirfarandi eru yfirferðir ásamt lyndisgildi þeirra, sem getur "
#     "verið 'jákvætt', 'hlutlaust' eða 'neikvætt'.",
#     prompt_template="Yfirferð: {text}\nLyndi: {label}",
#     prompt_label_mapping=dict(
#         positive="jákvætt", neutral="hlutlaust", negative="neikvætt"
#     ),
#     num_few_shot_examples=12,
#     max_generated_tokens=1,
# )

# FOREC_CONFIG = DatasetConfig(
#     name="forec",
#     pretty_name="the truncated version of FoReC",
#     huggingface_id="ScandEval/forec-mini",  # TODO: Needs to be uploaded
#     task=SENT,
#     languages=[FO],
#     prompt_prefix="Her koma nøkur ummæli og teirra kensluliga sjónarmið, sum kunnu "
#     "vera 'positivur', 'neutralur' ella 'negativur'.",
#     prompt_template="Ummæli: {text}\nKensluligt sjónarmið: {label}",
#     prompt_label_mapping=dict(
#         positive="positivur", neutral="neutralur", negative="negativur"
#     ),
#     num_few_shot_examples=12,
#     max_generated_tokens=1,
# )

SB10K_CONFIG = DatasetConfig(
    name="sb10k",
    pretty_name="the truncated version of the German sentiment classification "
    "dataset SB10k",
    huggingface_id="ScandEval/sb10k-mini",
    task=SENT,
    languages=[DE],
    prompt_prefix="Im Folgenden sind Tweets und ihre Stimmung aufgeführt, die "
    "'positiv', 'neutral' oder 'negativ' sein kann.",
    prompt_template="Tweet: {text}\nStimmungslage: {label}",
    prompt_label_mapping=dict(
        positive="positiv", neutral="neutral", negative="negativ"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

DUTCH_SOCIAL_CONFIG = DatasetConfig(
    name="dutch-social",
    pretty_name="the truncated version of the Dutch sentiment classification "
    "dataset Dutch Social",
    huggingface_id="ScandEval/dutch-social-mini",
    task=SENT,
    languages=[NL],
    prompt_prefix="Hieronder staan tweets en hun sentiment, dat 'positief', "
    "'neutraal' of 'negatief' kan zijn.",
    prompt_template="Tweet: {text}\nSentiment: {label}",
    prompt_label_mapping=dict(
        positive="positief", neutral="neutraal", negative="negatief"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SST5_CONFIG = DatasetConfig(
    name="sst5",
    pretty_name="the truncated version of the English sentiment classification "
    "dataset SST5",
    huggingface_id="ScandEval/sst5-mini",
    task=SENT,
    languages=[EN],
    prompt_prefix="The following are texts are their sentiment, which can be "
    "'positive', 'neutral' or 'negative'.",
    prompt_template="Text: {text}\nSentiment: {label}",
    prompt_label_mapping=dict(
        positive="positive", neutral="neutral", negative="negative"
    ),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

# TODO: Icelandic Sentiment Classification
# TODO: Faroese Sentiment Classification


### NAMED ENTITY RECOGNITION DATASETS ###

SUC3_CONFIG = DatasetConfig(
    name="suc3",
    pretty_name="the truncated version of the Swedish named entity recognition "
    "dataset SUC 3.0",
    huggingface_id="ScandEval/suc3-mini",
    task=NER,
    languages=[SV],
    prompt_prefix="Följande är meningar och JSON-ordböcker med de namngivna enheter "
    "som förekommer i den givna meningen.",
    prompt_template="Mening: {text}\nNamngivna entiteter: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "plats",
        "i-loc": "plats",
        "b-org": "organisation",
        "i-org": "organisation",
        "b-misc": "diverse",
        "i-misc": "diverse",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

DANSK_CONFIG = DatasetConfig(
    name="dansk",
    pretty_name="the truncated version of the Danish named entity recognition "
    "dataset DANSK",
    huggingface_id="ScandEval/dansk-mini",
    task=NER,
    languages=[DA],
    prompt_prefix="Følgende er sætninger og JSON-ordbøger med de navngivne enheder, "
    "som forekommer i den givne sætning.",
    prompt_template="Sætning: {text}\nNavngivne enheder: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "sted",
        "i-loc": "sted",
        "b-org": "organisation",
        "i-org": "organisation",
        "b-misc": "diverse",
        "i-misc": "diverse",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

NORNE_NB_CONFIG = DatasetConfig(
    name="norne-nb",
    pretty_name="the truncated version of the Bokmål part of the Norwegian named "
    "entity recognition dataset NorNE",
    huggingface_id="ScandEval/norne-nb-mini",
    task=NER,
    languages=[NB, NO],
    prompt_prefix="Følgende er fraser og JSON-ordbøker med de navngitte enhetene "
    "som forekommer i den gitte frasen.",
    prompt_template="Frase: {text}\nNavngitte enheter: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "sted",
        "i-loc": "sted",
        "b-org": "organisasjon",
        "i-org": "organisasjon",
        "b-misc": "diverse",
        "i-misc": "diverse",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

NORNE_NN_CONFIG = DatasetConfig(
    name="norne-nn",
    pretty_name="the truncated version of the Nynorsk part of the Norwegian named "
    "entity recognition dataset NorNE",
    huggingface_id="ScandEval/norne-nn-mini",
    task=NER,
    languages=[NN],
    prompt_prefix="Følgende er fraser og JSON-ordbøker med de navngitte enhetene "
    "som forekommer i den gitte frasen.",
    prompt_template="Frase: {text}\nNavngitte enheter: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "sted",
        "i-loc": "sted",
        "b-org": "organisasjon",
        "i-org": "organisasjon",
        "b-misc": "diverse",
        "i-misc": "diverse",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

MIM_GOLD_NER_CONFIG = DatasetConfig(
    name="mim-gold-ner",
    pretty_name="the truncated version of the Icelandic named entity recognition "
    "dataset MIM-GOLD-NER",
    huggingface_id="ScandEval/mim-gold-ner-mini",
    task=NER,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru setningar ásamt JSON lyklum með nefndum einingum "
    "sem koma fyrir í setningunum.",
    prompt_template="Setning: {text}\nNefndar einingar: {label}",
    prompt_label_mapping={
        "b-per": "einstaklingur",
        "i-per": "einstaklingur",
        "b-loc": "staðsetning",
        "i-loc": "staðsetning",
        "b-org": "stofnun",
        "i-org": "stofnun",
        "b-misc": "ýmislegt",
        "i-misc": "ýmislegt",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

FONE_CONFIG = DatasetConfig(
    name="fone",
    pretty_name="the truncated version of the Faroese named entity recognition "
    "dataset FoNE",
    huggingface_id="ScandEval/fone-mini",
    task=NER,
    languages=[FO],
    prompt_prefix="Her eru nakrir setningar og nakrar JSON orðabøkur við nevndar "
    "eindir, sum eru í setningunum.",
    prompt_template="Setningur: {text}\nNevndar eindir: {label}",
    prompt_label_mapping={
        "b-per": "persónur",
        "i-per": "persónur",
        "b-loc": "staður",
        "i-loc": "staður",
        "b-org": "felagsskapur",
        "i-org": "felagsskapur",
        "b-misc": "ymiskt",
        "i-misc": "ymiskt",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

GERMEVAL_CONFIG = DatasetConfig(
    name="germeval",
    pretty_name="the truncated version of the German named entity recognition "
    "dataset GermEval",
    huggingface_id="ScandEval/germeval-mini",
    task=NER,
    languages=[DE],
    prompt_prefix="Es folgen Sätze und JSON-Wörterbücher mit den benannten "
    "Entitäten, die in der angegebenen Phrase vorkommen.",
    prompt_template="Satz: {text}\nBenannte Entitäten: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "ort",
        "i-loc": "ort",
        "b-org": "organisation",
        "i-org": "organisation",
        "b-misc": "verschiedenes",
        "i-misc": "verschiedenes",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

CONLL_NL_CONFIG = DatasetConfig(
    name="conll-nl",
    pretty_name="the Dutch part of the truncated version of the named entity "
    "recognition dataset CoNLL 2002",
    huggingface_id="ScandEval/conll-nl-mini",
    task=NER,
    languages=[NL],
    prompt_prefix="Hieronder staan zinnen en JSON woordenboeken met de genoemde "
    "entiteiten die voorkomen in de gegeven zin.",
    prompt_template="Zin: {text}\nGenoemde entiteiten: {label}",
    prompt_label_mapping={
        "b-per": "persoon",
        "i-per": "persoon",
        "b-loc": "locatie",
        "i-loc": "locatie",
        "b-org": "organisatie",
        "i-org": "organisatie",
        "b-misc": "diversen",
        "i-misc": "diversen",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

CONLL_EN_CONFIG = DatasetConfig(
    name="conll-en",
    pretty_name="the truncated version of the English named entity recognition "
    "dataset CoNLL 2003",
    huggingface_id="ScandEval/conll-en-mini",
    task=NER,
    languages=[EN],
    prompt_prefix="Below are sentences and JSON dictionaries with the named "
    "entities that occur in the given sentence.",
    prompt_template="Sentence: {text}\nNamed entities: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "location",
        "i-loc": "location",
        "b-org": "organization",
        "i-org": "organization",
        "b-misc": "miscellaneous",
        "i-misc": "miscellaneous",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
)

DANE_CONFIG = DatasetConfig(
    name="dane",
    pretty_name="the truncated version of the Danish named entity recognition "
    "dataset DaNE",
    huggingface_id="ScandEval/dane-mini",
    task=NER,
    languages=[DA],
    prompt_prefix="Følgende er sætninger og JSON-ordbøger med de navngivne enheder, "
    "som forekommer i den givne sætning.",
    prompt_template="Sætning: {text}\nNavngivne enheder: {label}",
    prompt_label_mapping={
        "b-per": "person",
        "i-per": "person",
        "b-loc": "sted",
        "i-loc": "sted",
        "b-org": "organisation",
        "i-org": "organisation",
        "b-misc": "diverse",
        "i-misc": "diverse",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
    unofficial=True,
)

WIKIANN_FO_CONFIG = DatasetConfig(
    name="wikiann-fo",
    pretty_name="the truncated version of the Faroese part of the named entity "
    "recognition dataset WikiANN",
    huggingface_id="ScandEval/wikiann-fo-mini",
    task=NER,
    languages=[FO],
    prompt_prefix="Her eru nakrir setningar og nakrar JSON orðabøkur við nevndar "
    "eindir, sum eru í setningunum.",
    prompt_template="Setningur: {text}\nNevndar eindir: {label}",
    prompt_label_mapping={
        "b-per": "persónur",
        "i-per": "persónur",
        "b-loc": "staður",
        "i-loc": "staður",
        "b-org": "felagsskapur",
        "i-org": "felagsskapur",
        "b-misc": "ymiskt",
        "i-misc": "ymiskt",
    },
    num_few_shot_examples=8,
    max_generated_tokens=128,
    unofficial=True,
)


### LINGUISTIC ACCEPTABILITY DATASETS ###

SCALA_SV_CONFIG = DatasetConfig(
    name="scala-sv",
    pretty_name="The Swedish part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-sv",
    task=LA,
    languages=[SV],
    prompt_prefix="Följande är meningar och huruvida de är grammatiskt korrekta.",
    prompt_template="Mening: {text}\nGrammatisk korrekt: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nej"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_DA_CONFIG = DatasetConfig(
    name="scala-da",
    pretty_name="the Danish part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-da",
    task=LA,
    languages=[DA],
    prompt_prefix="Følgende er sætninger og om de er grammatisk korrekte.",
    prompt_template="Sætning: {text}\nGrammatisk korrekt: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nej"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_NB_CONFIG = DatasetConfig(
    name="scala-nb",
    pretty_name="the Bokmål part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-nb",
    task=LA,
    languages=[NB, NO],
    prompt_prefix="Følgende er setninger og hvorvidt de er grammatisk korrekte.",
    prompt_template="Setning: {text}\nGrammatisk korrekt: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nei"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_NN_CONFIG = DatasetConfig(
    name="scala-nn",
    pretty_name="the Nynorsk part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-nn",
    task=LA,
    languages=[NN],
    prompt_prefix="Følgende er setninger og hvorvidt de er grammatisk korrekte.",
    prompt_template="Setning: {text}\nGrammatisk korrekt: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nei"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_IS_CONFIG = DatasetConfig(
    name="scala-is",
    pretty_name="the Icelandic part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-is",
    task=LA,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru setningar og hvort þær eru málfræðilega réttar.",
    prompt_template="Setning: {text}\nMálfræðilega rétt: {label}",
    prompt_label_mapping=dict(correct="já", incorrect="nei"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)


SCALA_FO_CONFIG = DatasetConfig(
    name="scala-fo",
    pretty_name="the Faroese part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-fo",
    task=LA,
    languages=[FO],
    prompt_prefix="Hetta eru nakrir setningar og um teir eru mállæruliga rættir.",
    prompt_template="Setningur: {text}\nMállæruliga rættur: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nei"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_DE_CONFIG = DatasetConfig(
    name="scala-de",
    pretty_name="the German part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-de",
    task=LA,
    languages=[DE],
    prompt_prefix="Die folgenden Sätze und ob sie grammatikalisch korrekt sind.",
    prompt_template="Satz: {text}\nGrammatikalisch richtig: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nein"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_NL_CONFIG = DatasetConfig(
    name="scala-nl",
    pretty_name="the Dutch part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-nl",
    task=LA,
    languages=[NL],
    prompt_prefix="Hieronder staan zinnen en of ze grammaticaal correct zijn.",
    prompt_template="Zin: {text}\nGrammaticaal correct: {label}",
    prompt_label_mapping=dict(correct="ja", incorrect="nee"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)

SCALA_EN_CONFIG = DatasetConfig(
    name="scala-en",
    pretty_name="the English part of the linguistic acceptability dataset ScaLA",
    huggingface_id="ScandEval/scala-en",
    task=LA,
    languages=[EN],
    prompt_prefix="The following are sentences and whether they are grammatically "
    "correct.",
    prompt_template="Sentence: {text}\nGrammatically correct: {label}",
    prompt_label_mapping=dict(correct="yes", incorrect="no"),
    num_few_shot_examples=12,
    max_generated_tokens=1,
)


### EXTRACTIVE QUESTION ANSWERING DATASETS ###

SCANDIQA_DA_CONFIG = DatasetConfig(
    name="scandiqa-da",
    pretty_name="the Danish part of the truncated version of the question answering "
    "dataset ScandiQA",
    huggingface_id="ScandEval/scandiqa-da-mini",
    task=QA,
    languages=[DA],
    prompt_prefix="Følgende er tekster med tilhørende spørgsmål og svar.",
    prompt_template="Tekst: {text}\nSpørgsmål: {question}\nSvar med maks. 3 ord: "
    "{label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

NORQUAD_CONFIG = DatasetConfig(
    name="norquad",
    pretty_name="the truncated version of the Norwegian question answering "
    "dataset NorQuAD",
    huggingface_id="ScandEval/norquad-mini",
    task=QA,
    languages=[NB, NN, NO],
    prompt_prefix="Her følger tekster med tilhørende spørsmål og svar.",
    prompt_template="Tekst: {text}\nSpørsmål: {question}\nSvar på maks 3 ord: {label}",
    num_few_shot_examples=2,
    max_generated_tokens=32,
)

SCANDIQA_SV_CONFIG = DatasetConfig(
    name="scandiqa-sv",
    pretty_name="the Swedish part of the truncated version of the question answering "
    "dataset ScandiQA",
    huggingface_id="ScandEval/scandiqa-sv-mini",
    task=QA,
    languages=[SV],
    prompt_prefix="Nedan följer texter med tillhörande frågor och svar.",
    prompt_template="Text: {text}\nFråga: {question}\nSvar på max 3 ord: {label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

NQII_CONFIG = DatasetConfig(
    name="nqii",
    pretty_name="the truncated version of the Icelandic question answering dataset "
    "Natural Questions in Icelandic",
    huggingface_id="ScandEval/nqii-mini",
    task=QA,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru textar með tilheyrandi spurningum og svörum.",
    prompt_template="Texti: {text}\nSpurning: {question}\nSvaraðu með að hámarki 3 "
    "orðum: {label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

# FOQA_CONFIG = DatasetConfig(
#     name="foqa",
#     pretty_name="Faroese Question Answering",
#     huggingface_id="ScandEval/foqa-mini",  # TODO: Needs to be uploaded
#     task=QA,
#     languages=[FO],
#     prompt_template="{text}\nSpurningur: {question}\nSvara við í mesta lagi trimum "
#     "orðum: {label}",
#     num_few_shot_examples=4,
#     max_generated_tokens=32,
# )

GERMANQUAD_CONFIG = DatasetConfig(
    name="germanquad",
    pretty_name="the truncated version of the German question answering dataset "
    "GermanQuAD",
    huggingface_id="ScandEval/germanquad-mini",
    task=QA,
    languages=[DE],
    prompt_prefix="Im Folgenden finden Sie Texte mit den dazugehörigen Fragen und "
    "Antworten.",
    prompt_template="Text: {text}\nFragen: {question}\nFragen Antwort in maximal 3 "
    "Wörtern: {label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

SQUAD_CONFIG = DatasetConfig(
    name="squad",
    pretty_name="the truncated version of the English question answering "
    "dataset SQuAD",
    huggingface_id="ScandEval/squad-mini",
    task=QA,
    languages=[EN],
    prompt_prefix="The following are texts with accompanying questions and answers.",
    prompt_template="Text: {text}\nQuestion: {question}\nAnswer in max 3 words: "
    "{label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

SQUAD_NL_CONFIG = DatasetConfig(
    name="squad-nl",
    pretty_name="the truncated version of the Dutch question answering dataset "
    "SQuAD-nl, translated from the English SQuAD dataset",
    huggingface_id="ScandEval/squad-nl-mini",
    task=QA,
    languages=[NL],
    prompt_prefix="Hieronder volgen teksten met bijbehorende vragen en antwoorden.",
    prompt_template="Tekst: {text}\nVraag: {question}\nAntwoord in max 3 woorden: "
    "{label}",
    num_few_shot_examples=4,
    max_generated_tokens=32,
)

# TODO: Faroese Question Answering


### SUMMARIZATION DATASETS ###

NORDJYLLAND_NEWS_CONFIG = DatasetConfig(
    name="nordjylland-news",
    pretty_name="the truncated version of the Danish summarisation dataset "
    "Nordjylland News",
    huggingface_id="ScandEval/nordjylland-news-mini",
    task=SUMM,
    languages=[DA],
    prompt_prefix="Følgende er nyhedsartikler med tilhørende resuméer.",
    prompt_template="Nyhedsartikel: {text}\nResumé: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

MLSUM_CONFIG = DatasetConfig(
    name="mlsum",
    pretty_name="the truncated version of the German summarisation dataset MLSum",
    huggingface_id="ScandEval/mlsum-mini",
    task=SUMM,
    languages=[DE],
    prompt_prefix="Im Folgenden finden Sie Nachrichtenartikel mit den dazugehörigen "
    "Zusammenfassungen.",
    prompt_template="Nachrichtenartikel: {text}\nZusammenfassung: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

RRN_CONFIG = DatasetConfig(
    name="rrn",
    pretty_name="the truncated version of the Icelandic summarisation dataset "
    "RÚV Radio News",
    huggingface_id="ScandEval/rrn-mini",
    task=SUMM,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru fréttagreinar með tilheyrandi samantektum.",
    prompt_template="Fréttagrein: {text}\nSamantekt: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

NO_SAMMENDRAG_CONFIG = DatasetConfig(
    name="no-sammendrag",
    pretty_name="the truncated version of the Norwegian summarisation dataset "
    "Norske Sammendrag",
    huggingface_id="ScandEval/no-sammendrag-mini",
    task=SUMM,
    languages=[NB, NN, NO],
    prompt_prefix="Her følger nyhetsartikler med tilhørende sammendrag.",
    prompt_template="Nyhetsartikkel: {text}\nSammendrag: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

WIKI_LINGUA_NL_CONFIG = DatasetConfig(
    name="wiki-lingua-nl",
    pretty_name="the Dutch part of the truncated version of the summarisation dataset "
    "WikiLingua",
    huggingface_id="ScandEval/wiki-lingua-nl-mini",
    task=SUMM,
    languages=[NL],
    prompt_prefix="Hieronder volgen artikelen met bijbehorende samenvattingen.",
    prompt_template="Artikel: {text}\nSamenvatting: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

SWEDN_CONFIG = DatasetConfig(
    name="swedn",
    pretty_name="the truncated version of the Swedish summarisation dataset SweDN",
    huggingface_id="ScandEval/swedn-mini",
    task=SUMM,
    languages=[SV],
    prompt_prefix="Nedan följer artiklar med tillhörande sammanfattningar.",
    prompt_template="Artikel: {text}\nSammanfattning: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

CNN_DAILYMAIL_CONFIG = DatasetConfig(
    name="cnn-dailymail",
    pretty_name="the truncated version of the English summarisation dataset "
    "CNN-DailyMail",
    huggingface_id="ScandEval/cnn-dailymail-mini",
    task=SUMM,
    languages=[EN],
    prompt_prefix="The following are articles with accompanying summaries.",
    prompt_template="News article: {text}\nSummary: {target_text}",
    num_few_shot_examples=1,
    max_generated_tokens=256,
)

# TODO: Faroese summarization


### KNOWLEDGE DATASETS ###

DANSKE_TALEMAADER_CONFIG = DatasetConfig(
    name="danske-talemaader",
    pretty_name="the truncated version of the Danish knowledge dataset Danske "
    "Talemåder",
    huggingface_id="ScandEval/danske-talemaader-mini",
    task=KNOW,
    languages=[DA],
    prompt_prefix="Følgende er multiple choice spørgsmål (med svar).",
    prompt_template="Hvad er betydningen af følgende talemåde: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

DANISH_CITIZEN_TESTS_CONFIG = DatasetConfig(
    name="danish-citizen-tests",
    pretty_name="the Danish knowledge dataset Danish Citizen Tests",
    huggingface_id="ScandEval/danish-citizen-tests",
    task=KNOW,
    languages=[DA],
    prompt_prefix="Følgende er multiple choice spørgsmål (med svar).",
    prompt_template="Spørgsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_NO_CONFIG = DatasetConfig(
    name="mmlu-no",
    pretty_name="the truncated version of the Norwegian knowledge dataset MMLU-no, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-no-mini",
    task=KNOW,
    languages=[NB, NN, NO],
    prompt_prefix="Følgende er flervalgsspørsmål (med svar).",
    prompt_template="Spørsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_SV_CONFIG = DatasetConfig(
    name="mmlu-sv",
    pretty_name="the truncated version of the Swedish knowledge dataset MMLU-sv, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-sv-mini",
    task=KNOW,
    languages=[SV],
    prompt_prefix="Följande är flervalsfrågor (med svar).",
    prompt_template="Fråga: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_IS_CONFIG = DatasetConfig(
    name="mmlu-is",
    pretty_name="the truncated version of the Icelandic knowledge dataset MMLU-is, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-is-mini",
    task=KNOW,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru fjölvalsspurningar (með svörum).",
    prompt_template="Spurningar: {text}\nSvara: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_DE_CONFIG = DatasetConfig(
    name="mmlu-de",
    pretty_name="the truncated version of the German knowledge dataset MMLU-de, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-de-mini",
    task=KNOW,
    languages=[DE],
    prompt_prefix="Die folgenden Fragen sind Multiple-Choice-Fragen (mit Antworten).",
    prompt_template="Frage: {text}\nAntwort: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_NL_CONFIG = DatasetConfig(
    name="mmlu-nl",
    pretty_name="the truncated version of the Dutch knowledge dataset MMLU-nl, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-nl-mini",
    task=KNOW,
    languages=[NL],
    prompt_prefix="Hieronder staan meerkeuzevragen (met antwoorden).",
    prompt_template="Vraag: {text}\nAntwoord: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_CONFIG = DatasetConfig(
    name="mmlu",
    pretty_name="the truncated version of the English knowledge dataset MMLU",
    huggingface_id="ScandEval/mmlu-mini",
    task=KNOW,
    languages=[EN],
    prompt_prefix="The following are multiple choice questions (with answers).",
    prompt_template="Question: {text}\nAnswer: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

MMLU_DA_CONFIG = DatasetConfig(
    name="mmlu-da",
    pretty_name="the truncated version of the Danish knowledge dataset MMLU-da, "
    "translated from the English MMLU dataset",
    huggingface_id="ScandEval/mmlu-da-mini",
    task=KNOW,
    languages=[DA],
    prompt_prefix="Følgende er multiple choice spørgsmål (med svar).",
    prompt_template="Spørgsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_DA_CONFIG = DatasetConfig(
    name="arc-da",
    pretty_name="the truncated version of the Danish knowledge dataset ARC-da, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-da-mini",
    task=KNOW,
    languages=[DA],
    prompt_prefix="Følgende er multiple choice spørgsmål (med svar).",
    prompt_template="Spørgsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_NO_CONFIG = DatasetConfig(
    name="arc-no",
    pretty_name="the truncated version of the Norwegian knowledge dataset ARC-no, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-no-mini",
    task=KNOW,
    languages=[NB, NN, NO],
    prompt_prefix="Følgende er flervalgsspørsmål (med svar).",
    prompt_template="Spørsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_SV_CONFIG = DatasetConfig(
    name="arc-sv",
    pretty_name="the truncated version of the Swedish knowledge dataset ARC-sv, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-sv-mini",
    task=KNOW,
    languages=[SV],
    prompt_prefix="Följande är flervalsfrågor (med svar).",
    prompt_template="Fråga: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_IS_CONFIG = DatasetConfig(
    name="arc-is",
    pretty_name="the truncated version of the Icelandic knowledge dataset ARC-is, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-is-mini",
    task=KNOW,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru fjölvalsspurningar (með svörum).",
    prompt_template="Spurningar: {text}\nSvara: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_DE_CONFIG = DatasetConfig(
    name="arc-de",
    pretty_name="the truncated version of the German knowledge dataset ARC-de, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-de-mini",
    task=KNOW,
    languages=[DE],
    prompt_prefix="Die folgenden Fragen sind Multiple-Choice-Fragen (mit Antworten).",
    prompt_template="Frage: {text}\nAntwort: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_NL_CONFIG = DatasetConfig(
    name="arc-nl",
    pretty_name="the truncated version of the Dutch knowledge dataset ARC-nl, "
    "translated from the English ARC dataset",
    huggingface_id="ScandEval/arc-nl-mini",
    task=KNOW,
    languages=[NL],
    prompt_prefix="Hieronder staan meerkeuzevragen (met antwoorden).",
    prompt_template="Vraag: {text}\nAntwoord: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

ARC_CONFIG = DatasetConfig(
    name="arc",
    pretty_name="the truncated version of the English knowledge dataset ARC",
    huggingface_id="ScandEval/arc-mini",
    task=KNOW,
    languages=[EN],
    prompt_prefix="The following are multiple choice questions (with answers).",
    prompt_template="Question: {text}\nAnswer: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

# TODO: Faroese knowledge


### COMMON SENSE REASONING DATASETS ###

HELLASWAG_DA_CONFIG = DatasetConfig(
    name="hellaswag-da",
    pretty_name="the truncated version of the Danish common-sense reasoning dataset "
    "HellaSwag-da, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-da-mini",
    task=COMMON_SENSE,
    languages=[DA],
    prompt_prefix="Følgende er multiple choice spørgsmål (med svar).",
    prompt_template="Spørgsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_NO_CONFIG = DatasetConfig(
    name="hellaswag-no",
    pretty_name="the truncated version of the Norwegian common-sense reasoning dataset "
    "HellaSwag-no, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-no-mini",
    task=COMMON_SENSE,
    languages=[NB, NN, NO],
    prompt_prefix="Følgende er flervalgsspørsmål (med svar).",
    prompt_template="Spørsmål: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_SV_CONFIG = DatasetConfig(
    name="hellaswag-sv",
    pretty_name="the truncated version of the Swedish common-sense reasoning dataset "
    "HellaSwag-sv, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-sv-mini",
    task=COMMON_SENSE,
    languages=[SV],
    prompt_prefix="Följande är flervalsfrågor (med svar).",
    prompt_template="Fråga: {text}\nSvar: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_IS_CONFIG = DatasetConfig(
    name="hellaswag-is",
    pretty_name="the truncated version of the Icelandic common-sense reasoning dataset "
    "HellaSwag-is, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-is-mini",
    task=COMMON_SENSE,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru fjölvalsspurningar (með svörum).",
    prompt_template="Spurningar: {text}\nSvara: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
    unofficial=True,
)

WINOGRANDE_IS = DatasetConfig(
    name="winogrande-is",
    pretty_name="the Icelandic common-sense reasoning dataset "
    "Winogrande-is, manually translated from the English Winogrande dataset",
    huggingface_id="ScandEval/winogrande-is",
    task=COMMON_SENSE,
    languages=[IS],
    prompt_prefix="Eftirfarandi eru fjölvalsspurningar (með svörum).",
    prompt_template="Spurningar: {text}\nSvara: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_DE_CONFIG = DatasetConfig(
    name="hellaswag-de",
    pretty_name="the truncated version of the German common-sense reasoning dataset "
    "HellaSwag-de, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-de-mini",
    task=COMMON_SENSE,
    languages=[DE],
    prompt_prefix="Die folgenden Fragen sind Multiple-Choice-Fragen (mit Antworten).",
    prompt_template="Frage: {text}\nAntwort: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_NL_CONFIG = DatasetConfig(
    name="hellaswag-nl",
    pretty_name="the truncated version of the Dutch common-sense reasoning dataset "
    "HellaSwag-nl, translated from the English HellaSwag dataset",
    huggingface_id="ScandEval/hellaswag-nl-mini",
    task=COMMON_SENSE,
    languages=[NL],
    prompt_prefix="Hieronder staan meerkeuzevragen (met antwoorden).",
    prompt_template="Vraag: {text}\nAntwoord: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

HELLASWAG_CONFIG = DatasetConfig(
    name="hellaswag",
    pretty_name="the truncated version of the English common-sense reasoning "
    "dataset HellaSwag",
    huggingface_id="ScandEval/hellaswag-mini",
    task=COMMON_SENSE,
    languages=[EN],
    prompt_prefix="The following are multiple choice questions (with answers).",
    prompt_template="Question: {text}\nAnswer: {label}",
    prompt_label_mapping=dict(a="a", b="b", c="c", d="d"),
    num_few_shot_examples=5,
    max_generated_tokens=1,
)

# TODO: Faroese common sense reasoning


### SPEED ESTIMATION DATASETS ###

SPEED_CONFIG = DatasetConfig(
    name="speed",
    pretty_name="the speed estimation benchmark",
    huggingface_id="",
    task=SPEED,
    languages=list(get_all_languages().values()),
    prompt_template="",
    max_generated_tokens=1,
)



"""Text-to-text generation benchmark dataset."""

import logging
import random
from typing import TYPE_CHECKING, Any

import numpy as np
from transformers.data.data_collator import DataCollatorWithPadding

from .benchmark_dataset import BenchmarkDataset
from .exceptions import InvalidBenchmark
from .generation import extract_raw_predictions
from .utils import (
    METRIC_ATTRIBUTES_TAKING_UP_MEMORY,
    HiddenPrints,
    clear_memory,
    raise_if_model_output_contains_nan_values,
)

if TYPE_CHECKING:
    from datasets.arrow_dataset import Dataset
    from transformers import BatchEncoding, PreTrainedModel
    from transformers.utils import ModelOutput

    from .protocols import GenerativeModel, Tokenizer
    from .types import Labels, Predictions


logger = logging.getLogger(__package__)


class TextToText(BenchmarkDataset):
    """Text-to-text benchmark dataset.

    Args:
        dataset_config:
            The dataset configuration.
        benchmark_config:
            The benchmark configuration.

    Attributes:
        dataset_config:
            The configuration of the dataset.
        benchmark_config:
            The configuration of the benchmark.
    """

    def _preprocess_data(self, dataset: "Dataset", **kwargs) -> "Dataset":
        """Preprocess a dataset by tokenizing and aligning the labels.

        Args:
            dataset:
                The dataset to preprocess.
            kwargs:
                Extra keyword arguments containing objects used in preprocessing the
                dataset.

        Returns:
            The preprocessed dataset.
        """
        tokenizer: "Tokenizer" = kwargs["tokenizer"]

        def tokenise(examples: dict) -> "BatchEncoding":
            return tokenizer(text=examples["text"], truncation=True, padding=False)

        tokenised = dataset.map(
            tokenise, batched=True, load_from_cache_file=False, keep_in_memory=True
        )

        return tokenised

    def _load_data_collator(
        self,
        tokenizer: "Tokenizer | None" = None,
        model: "PreTrainedModel | GenerativeModel | None" = None,
    ):
        """Load the data collator used to prepare samples during finetuning.

        Args:
            tokenizer:
                A pretrained tokenizer. Can be None if the tokenizer is not used in the
                initialisation of the data collator. Defaults to None.
            model:
                A pretrained model. Can be None if the model is not used. Defaults to
                None.

        Returns:
            The data collator.
        """
        return DataCollatorWithPadding(tokenizer, padding="longest")

    def _compute_metrics(
        self,
        model_outputs_and_labels: tuple["Predictions", "Labels"],
        id2label: dict[int, str],
    ) -> dict[str, float]:
        """Compute the metrics needed for evaluation.

        Args:
            model_outputs_and_labels:
                The first sequence contains the model outputs and the second sequence
                contains the true labels.
            id2label:
                Conversion of indices to labels.

        Returns:
            A dictionary with the names of the metrics as keys and the metric values as
            values.
        """
        model_outputs, labels = model_outputs_and_labels

        raise_if_model_output_contains_nan_values(model_output=model_outputs)

        model_output_dtype = np.asarray(model_outputs).dtype
        output_is_prob = model_output_dtype in [np.float16, np.float32, np.float64]
        if output_is_prob:
            predictions = np.asarray(model_outputs).argmax(axis=-1)
        else:
            predictions = model_outputs

        results: dict[str, float] = dict()
        for cfg in self.dataset_config.task.metrics:
            metric = self._metrics[cfg.name]

            # Some metrics can be computed on hardware accelerators. In this case we
            # start by setting the device to the same device as the model
            if cfg.compute_kwargs.get("device", None) == "auto":
                cfg.compute_kwargs["device"] = self.benchmark_config.device.type

            while True:
                try:
                    with HiddenPrints():
                        score_dict: dict[str, float] | None = metric.compute(
                            predictions=predictions,
                            references=labels,
                            **cfg.compute_kwargs,
                        )

                    # Clear the cache of the BERTScorer to avoid memory leaks
                    for attribute in METRIC_ATTRIBUTES_TAKING_UP_MEMORY:
                        if hasattr(metric, attribute):
                            delattr(metric, attribute)

                    clear_memory()
                    break
                except Exception as e:
                    # Clear the cache of the BERTScorer to avoid memory leaks
                    if hasattr(metric, "cached_bertscorer"):
                        del metric.cached_bertscorer
                        clear_memory()

                    oom_error = [
                        "CUDA out of memory",
                        "CUDA error",
                        "MPS backend out of memory",
                    ]
                    if not any(error in str(e) for error in oom_error):
                        raise InvalidBenchmark(str(e))

                    if cfg.compute_kwargs.get("batch_size", 1) > 1:
                        batch_size = cfg.compute_kwargs["batch_size"]
                        cfg.compute_kwargs["batch_size"] = batch_size // 2
                        logger.debug(
                            "Out of memory error occurred during the computation of "
                            f"the metric {cfg.pretty_name}. Reducing the batch size to "
                            f"{cfg.compute_kwargs['batch_size']}."
                        )
                    elif cfg.compute_kwargs.get("device", "cpu") != "cpu":
                        cfg.compute_kwargs["batch_size"] = 32
                        cfg.compute_kwargs["device"] = "cpu"
                        logger.debug(
                            "Out of memory error occurred during the computation of "
                            f"the metric {cfg.pretty_name}. Moving the computation to "
                            "the CPU."
                        )
                    else:
                        raise InvalidBenchmark(str(e))

            # The metric returns None if we are running on multi-GPU and the current
            # process is not the main process
            if score_dict is not None:
                scores = score_dict[cfg.results_key]
                if isinstance(scores, list):
                    scores = sum(scores) / len(scores)
                results[cfg.name] = scores

        return results

    def _extract_few_shot_examples(
        self, train_dataset: "Dataset", random_seed: int
    ) -> list[dict[str, Any]]:
        """Extract few-shot examples from the training dataset.

        Args:
            train_dataset:
                The training dataset.
            random_seed:
                The random seed to use when extracting the few-shot examples.

        Returns:
            The few-shot examples.
        """
        shuffled_train = train_dataset.shuffle(seed=random_seed)
        num_few_shots = self.dataset_config.num_few_shot_examples
        few_shot_examples: list[dict[str, Any]] = list()

        # We pick the few-shot examples one at a time rather than all at once since
        # we're working with a bootstrapped training dataset, meaning that it will have
        # duplicates. This ensures that we don't have any duplicates in the few-shot
        # examples
        while len(few_shot_examples) < num_few_shots:
            example = shuffled_train.select(range(1))[0]
            few_shot_examples.append(example)
            shuffled_train = shuffled_train.filter(
                lambda x: x["text"] != example["text"]
            )

        random.seed(random_seed)
        random.shuffle(few_shot_examples)
        return few_shot_examples

    def _apply_few_shot_prompt(
        self, examples: dict, few_shot_examples: list[dict], tokenizer: "Tokenizer"
    ) -> dict:
        """Apply a few-shot prompt to the examples.

        Args:
            examples:
                The examples to apply the prompt to.
            few_shot_examples:
                The examples to be included in the few-shot prompt.
            tokenizer:
                The tokenizer to use to encode the few-shot prompt.

        Returns:
            The examples with the few-shot prompt applied.
        """
        # Build the few-shot part of the prompt
        few_shot_prompts = [
            self.dataset_config.prompt_template.format(
                text=example["text"].replace("\n", " ").strip(),
                target_text=example["target_text"].replace("\n", " ").strip(),
            )
            for example in few_shot_examples
        ]
        prompt_prefix = ""
        if self.dataset_config.prompt_prefix:
            prompt_prefix = self.dataset_config.prompt_prefix + "\n\n"
        few_shot_prompt = prompt_prefix + "\n\n".join(few_shot_prompts)

        # Add the texts from the examples to the prompts. We remove newlines from the
        # examples as they have the special function to separate the few-shot examples
        # from one another
        new_prompts = [
            self.dataset_config.prompt_template.format(
                text=text.replace("\n", " ").strip(), target_text=""
            ).strip()
            for text in examples["text"]
        ]

        final_prompts = [
            few_shot_prompt + "\n\n" + new_prompt for new_prompt in new_prompts
        ]

        examples["text"] = final_prompts

        return examples

    def _extract_labels_from_generation(
        self,
        input_batch: dict[str, list],
        model_output: "ModelOutput",
        tokenizer: "Tokenizer",
    ) -> list[Any]:
        """Extract the predicted labels from the generated output.

        Args:
            input_batch:
                The input batch, where the keys are the feature names and the values
                are lists with the feature values.
            model_output:
                The raw generated output of the model.
            tokenizer:
                The tokenizer used together with the model.

        Returns:
            The predicted labels.
        """
        raw_predictions = extract_raw_predictions(
            generated_sequences=model_output.sequences, tokenizer=tokenizer
        )
        return raw_predictions



"""Benchmarking model inference speed."""

import logging
from collections import defaultdict
from typing import TYPE_CHECKING

import pyinfer
import torch
from tqdm.auto import tqdm
from transformers.modeling_utils import GenerationConfig, PreTrainedModel

from .exceptions import InvalidBenchmark
from .model_loading import load_model
from .utils import clear_memory, model_is_generative

if TYPE_CHECKING:
    from transformers.tokenization_utils import PreTrainedTokenizer

    from .config import BenchmarkConfig, DatasetConfig, ModelConfig
    from .protocols import GenerativeModel, Tokenizer


logger = logging.getLogger(__package__)


def benchmark_speed(
    itr: tqdm,
    tokenizer: "PreTrainedTokenizer",
    model: PreTrainedModel,
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
) -> dict[str, list[dict[str, float]]]:
    """Benchmark model inference speed.

    Args:
        itr:
            tqdm iterator.
        tokenizer:
            Tokenizer to use.
        model:
            Model to use.
        model_config:
            Model configuration.
        dataset_config:
            Dataset configuration.
        benchmark_config:
            Benchmark configuration.

    Returns:
        Dictionary of scores.
    """
    scores: dict[str, list[dict[str, float]]] = defaultdict(list)

    for itr_idx in itr:
        itr_scores = benchmark_speed_single_iteration(
            tokenizer=tokenizer,
            model=model,
            itr_idx=itr_idx,
            model_config=model_config,
            dataset_config=dataset_config,
            benchmark_config=benchmark_config,
        )
        clear_memory()

        if isinstance(itr_scores, Exception):
            raise InvalidBenchmark(f"Speed benchmark failed with error: {itr_scores!r}")
        else:
            scores["test"].append(itr_scores["test"])
            if benchmark_config.evaluate_train:
                scores["train"].append(itr_scores["train"])
            logger.debug(f"Scores for iteration {itr_idx}: {itr_scores}")

    return scores


def benchmark_speed_single_iteration(
    tokenizer: "Tokenizer",
    model: "PreTrainedModel | GenerativeModel",
    itr_idx: int,
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
) -> dict[str, dict[str, float]] | Exception:
    """Run a single iteration of the speed benchmark.

    Args:
        tokenizer:
            The tokenizer to use in the benchmark.
        model:
            The model to use in the benchmark.
        itr_idx:
            The index of the iteration.
        model_config:
            The model configuration.
        dataset_config:
            The dataset configuration.
        benchmark_config:
            The benchmark configuration.

    Returns:
        A dictionary containing the scores for the current iteration, with keys `train`
        and `test`. If an exception is raised, then the exception is returned.
    """
    is_generative = model_is_generative(model=model)

    scores: dict[str, dict[str, float]] = dict()
    try:
        # Reinitialise a new model
        if tokenizer is None or model is None:
            model, tokenizer = load_model(
                model_config=model_config,
                dataset_config=dataset_config,
                benchmark_config=benchmark_config,
            )

        def predict(doc: str) -> None:
            """Function used to benchmark inference speed of the model."""
            # Raise an error if the tokenizer or model is undefined
            if tokenizer is None or model is None:
                raise ValueError("Tokenizer and model must not be None.")

            # Tokenize the document
            inputs = tokenizer(doc, padding=True, truncation=True, return_tensors="pt")

            inputs = {name: tensor.to(model.device) for name, tensor in inputs.items()}

            # Run inference with the model
            with torch.inference_mode():
                if is_generative:
                    model.generate(
                        inputs=inputs["input_ids"],
                        generation_config=GenerationConfig(
                            max_new_tokens=1,
                            pad_token_id=model.config.pad_token_id,
                            eos_token_id=model.config.eos_token_id,
                            do_sample=False,
                        ),
                    )
                else:
                    assert isinstance(model, PreTrainedModel)
                    model(**inputs)

        base_doc = "Document which contains roughly 10 tokens. "
        multiplier = 10 * (1 + itr_idx)
        doc = base_doc * multiplier
        short_multiplier = 1.25 * (1 + itr_idx)
        short_doc = base_doc * round(short_multiplier)

        # Do a warmup run, as the first run is always slower
        pyinfer.InferenceReport(model=predict, inputs=base_doc, n_seconds=1).run(
            print_report=False
        )

        speed_scores = pyinfer.InferenceReport(
            model=predict, inputs=doc, n_seconds=3
        ).run(print_report=False)
        num_tokens = len(tokenizer([doc], truncation=True)["input_ids"][0])
        tokens_per_second = speed_scores["Infer(p/sec)"] * num_tokens

        speed_scores_short = pyinfer.InferenceReport(
            model=predict, inputs=short_doc, n_seconds=3
        ).run(print_report=False)
        num_tokens_short = len(tokenizer([short_doc], truncation=True)["input_ids"][0])
        tokens_per_second_short = speed_scores_short["Infer(p/sec)"] * num_tokens_short

        scores["test"] = dict(
            test_speed=tokens_per_second, test_speed_short=tokens_per_second_short
        )
        if benchmark_config.evaluate_train:
            scores["train"] = dict(
                train_speed=tokens_per_second, train_speed_short=tokens_per_second_short
            )

        # Return the scores
        return scores

    except (RuntimeError, ValueError, IndexError) as e:
        return e



"""Protocols used throughout the project."""

from typing import TYPE_CHECKING, Literal, Protocol, Union, runtime_checkable

if TYPE_CHECKING:
    from torch import Tensor, device
    from transformers import (
        BatchEncoding,
        GenerationConfig,
        PretrainedConfig,
        PreTrainedModel,
    )
    from transformers.utils import ModelOutput

    from .config import BenchmarkConfig, DatasetConfig, ModelConfig


@runtime_checkable
class Tokenizer(Protocol):
    """A protocol for a tokenizer."""

    cls_token: str
    sep_token: str
    bos_token: str
    eos_token: str
    pad_token: str
    unk_token: str
    cls_token_id: int
    sep_token_id: int
    bos_token_id: int
    eos_token_id: int
    pad_token_id: int
    unk_token_id: int
    is_fast: bool
    chat_template: str | None

    def __call__(self, text: str | list[str], **kwargs) -> "BatchEncoding":
        """Call the tokenizer.

        Args:
            text:
                The text to tokenize.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The encoded inputs.
        """
        ...

    def decode(self, token_ids: list[int], **kwargs) -> str:
        """Decode a list of token IDs.

        Args:
            token_ids:
                The token IDs to decode.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The decoded string.
        """
        ...

    def batch_decode(self, sequences: list[list[int]], **kwargs) -> list[str]:
        """Decode a batch of token IDs.

        Args:
            sequences:
                The token IDs to decode.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The decoded strings.
        """
        ...

    def encode(self, text: str | list[str] | list[int], **kwargs) -> list[int]:
        """Encode one or more texts.

        Args:
            text:
                The text(s) to encode.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The encoded token IDs.
        """
        ...

    def convert_ids_to_tokens(
        self, ids: int | list[int], skip_special_tokens: bool = False
    ) -> str | list[str]:
        """Convert a list of token IDs to tokens.

        Args:
            ids:
                The token IDs to convert.
            skip_special_tokens:
                Whether to skip special tokens.

        Returns:
            The tokens.
        """
        ...

    def convert_tokens_to_ids(self, tokens: str | list[str]) -> int | list[int]:
        """Convert a list of tokens to token IDs.

        Args:
            tokens:
                The tokens to convert.

        Returns:
            The token IDs.
        """
        ...

    @property
    def special_tokens_map(self) -> dict[str, str | list[str]]:
        """The mapping from special tokens to their token strings."""
        ...

    @property
    def model_max_length(self) -> int:
        """The maximum length of a sequence that can be processed by the model."""
        ...

    @property
    def vocab_size(self) -> int:
        """Return the size of the vocabulary."""
        ...

    def pad(
        self,
        encoded_inputs: Union[
            "BatchEncoding",
            list["BatchEncoding"],
            dict[str, list[str]],
            dict[str, list[list[str]]],
            list[dict[str, list[str]]],
        ],
        **kwargs,
    ) -> "BatchEncoding":
        """Pad a batch of encoded inputs.

        Args:
            encoded_inputs:
                The encoded inputs to pad.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The padded encoded inputs.
        """
        ...

    def apply_chat_template(
        self, conversation: list[dict[Literal["role", "content"], str]], **kwargs
    ) -> str | list[int]:
        """Apply a chat template to a conversation.

        Args:
            conversation:
                The conversation.
            **kwargs:
                Keyword arguments to pass to the tokenizer.

        Returns:
            The conversation as a string.
        """
        ...


@runtime_checkable
class GenerativeModel(Protocol):
    """A protocol for a generative model."""

    @property
    def config(self) -> "PretrainedConfig":
        """The Hugging Face model configuration."""
        ...

    @property
    def device(self) -> "device":
        """The device on which the model is running."""
        ...

    def generate(
        self,
        inputs: "Tensor",
        generation_config: "GenerationConfig | None" = None,
        **generation_kwargs,
    ) -> "ModelOutput | Tensor":
        """Generate text.

        Args:
            inputs:
                The input IDs.
            generation_config:
                The generation configuration.
            **generation_kwargs:
                Keyword arguments to pass to the generation method.

        Returns:
            The generated text.
        """
        ...


class ModelSetup(Protocol):
    """A protocol for a general model setup."""

    def __init__(self, benchmark_config: "BenchmarkConfig") -> None:
        """Initialize the model setup.

        Args:
            benchmark_config:
                The benchmark configuration.
        """
        ...

    def model_exists(self, model_id: str) -> bool | dict[str, str]:
        """Check whether a model exists.

        Args:
            model_id:
                The model ID.

        Returns:
            Whether the model exist, or a dictionary explaining why we cannot check
            whether the model exists.
        """
        ...

    def get_model_config(self, model_id: str) -> "ModelConfig":
        """Get the model configuration.

        Args:
            model_id:
                The model ID.

        Returns:
            The model configuration.
        """
        ...

    def load_model(
        self, model_config: "ModelConfig", dataset_config: "DatasetConfig"
    ) -> tuple["PreTrainedModel | GenerativeModel", "Tokenizer"]:
        """Load a model.

        Args:
            model_config:
                The model configuration.
            dataset_config:
                The dataset configuration.

        Returns:
            The tokenizer and model.
        """
        ...



"""Model setup for OpenAI models."""

import importlib.util
import logging
import re
from typing import TYPE_CHECKING

from transformers import PretrainedConfig

from ..config import ModelConfig
from ..enums import Framework, ModelType
from ..openai_models import OpenAIModel, OpenAITokenizer
from ..utils import create_model_cache_dir

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from ..config import BenchmarkConfig, DatasetConfig
    from ..protocols import GenerativeModel, Tokenizer

if importlib.util.find_spec("openai") is not None:
    import openai

    # Older versions of `openai` doesn't have the `models` module, so we need to check
    # that, as it will cause errors later otherwise
    openai.models


logger = logging.getLogger(__package__)


# This is a list of the major models that OpenAI has released
CACHED_OPENAI_MODEL_IDS: list[str] = [
    "ada|babbage|curie|davinci",
    "(code|text)-(ada|babbage|curie|davinci)-[0-9]{3}",
    "gpt-3.5-turbo(-16k|-instruct)?(-[0-9]{4})?(-preview)?",
    "gpt-4(-[0-9]{4})?(-preview)?",
    "gpt-4-turbo(-[0-9]{4}-[0-9]{2}-[0-9]{2})?",
    "gpt-4-32k(-[0-9]{4})?(-preview)?",
]


VOCAB_SIZE_MAPPING = {
    "(text-)?(ada|babbage|curie|davinci)(-001)?": 50_257,
    "(code|text)-davinci-00[2-9]": 50_281,
    "gpt-3.5-turbo(-16k)?(-[0-9]{4})?": 100_256,
    "gpt-4-(32k)?(-[0-9]{4})?": 100_256,
    "gpt-4-[0-9]{4}-preview": 100_256,
    "gpt-4-turbo(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": 100_256,
    "gpt-4-(vision|turbo)(-preview)?": 100_256,
    "gpt-3.5-turbo-instruct(-[0-9]{4})?": 100_256,
    "gpt-4o(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": 200_019,
}


MODEL_MAX_LENGTH_MAPPING = {
    "(text-)?(ada|babbage|curie|davinci)(-001)?": 2_050,
    "text-davinci-00[2-9]": 4_098,
    "code-davinci-00[1-9]": 8_002,
    "gpt-3.5-turbo-0613": 4_096,
    "gpt-3.5-turbo(-[0-9]{4})?": 16_385,
    "gpt-3.5-turbo-16k(-[0-9]{4})?": 16_384,
    "gpt-4(-[0-9]{4})?": 8_191,
    "gpt-4-32k(-[0-9]{4})?": 32_767,
    "gpt-4-[0-9]{4}-preview": 128_000,
    "gpt-4-turbo(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": 128_000,
    "gpt-4-(vision|turbo)(-preview)?": 128_000,
    "gpt-3.5-turbo-instruct(-[0-9]{4})?": 4_095,
    "gpt-4o(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": 128_000,
}


NUM_PARAMS_MAPPING = {
    "(text-)?ada(-001)?": 350_000_000,
    "(text-)?babbage(-001)?": 3_000_000_000,
    "(text-)?curie(-001)?": 13_000_000_000,
    "((text|code)-)?davinci(-00[1-9])?": 175_000_000_000,
    "gpt-(3.5|4)-turbo-((16|32)k)?(-[0-9]{4})?": -1,
    "gpt-4-[0-9]{4}-preview": -1,
    "gpt-4-turbo(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": -1,
    "gpt-4-(vision|turbo)(-preview)?": -1,
    "gpt-3.5-turbo-instruct(-[0-9]{4})?": -1,
    "gpt-4o(-[0-9]{4}-[0-9]{2}-[0-9]{2})?": -1,
}


class OpenAIModelSetup:
    """Model setup for OpenAI models.

    Attributes:
        benchmark_config:
            The benchmark configuration.
    """

    def __init__(self, benchmark_config: "BenchmarkConfig") -> None:
        """Initialize the model setup.

        Args:
            benchmark_config:
                The benchmark configuration.
        """
        self.benchmark_config = benchmark_config

    def model_exists(self, model_id: str) -> bool | dict[str, str]:
        """Check if a model ID denotes an OpenAI model.

        Args:
            model_id:
                The model ID.

        Returns:
            Whether the model exists, or a dictionary explaining why we cannot check
            whether the model exists.
        """
        if importlib.util.find_spec("openai") is None:
            return dict(missing_extra="openai")

        # The model ID for the Azure OpenAI API is the deployment name and therefore
        # different from the model ID used in the OpenAI API. We'll just assume that
        # the model exists in this case.
        if self.benchmark_config.azure_openai_api_key is not None:
            return True

        all_models: list[openai.models.Model] = list()
        try:
            all_models = list(openai.models.list())
        except openai.OpenAIError as e:
            model_exists = any(
                [
                    re.match(pattern=model_pattern, string=model_id) is not None
                    for model_pattern in CACHED_OPENAI_MODEL_IDS
                ]
            )
            if not model_exists:
                if "OPENAI_API_KEY" in str(e):
                    return dict(missing_env_var="OPENAI_API_KEY")
                elif "AZURE_OPENAI_API_KEY" in str(e):
                    return dict(missing_env_var="AZURE_OPENAI_API_KEY")
                elif "AZURE_OPENAI_ENDPOINT" in str(e):
                    return dict(missing_env_var="AZURE_OPENAI_ENDPOINT")

        return model_id in [model.id for model in all_models]

    def get_model_config(self, model_id: str) -> ModelConfig:
        """Fetches configuration for an OpenAI model.

        Args:
            model_id:
                The model ID of the model.

        Returns:
            The model configuration.
        """
        return ModelConfig(
            model_id=model_id,
            revision="main",
            framework=Framework.API,
            task="text-generation",
            languages=list(),
            model_type=ModelType.OPENAI,
            model_cache_dir=create_model_cache_dir(
                cache_dir=self.benchmark_config.cache_dir, model_id=model_id
            ),
        )

    def load_model(
        self, model_config: ModelConfig, dataset_config: "DatasetConfig"
    ) -> tuple["PreTrainedModel | GenerativeModel", "Tokenizer"]:
        """Load an OpenAI model.

        Args:
            model_config:
                The model configuration.
            dataset_config:
                The dataset configuration.

        Returns:
            The tokenizer and model.
        """
        hf_model_config = PretrainedConfig.from_pretrained("gpt2")

        vocab_sizes = [
            vocab_size
            for pattern, vocab_size in VOCAB_SIZE_MAPPING.items()
            if re.match(pattern=pattern, string=model_config.model_id)
        ]
        hf_model_config.vocab_size = vocab_sizes[0] if vocab_sizes else 100_256

        # We subtract the maximum generation length, as that counts towards the total
        # amount of tokens that the model needs to process.
        # We subtract 1 as errors occur if the model is exactly at the maximum length.
        model_lengths = [
            model_length - dataset_config.max_generated_tokens - 1
            for pattern, model_length in MODEL_MAX_LENGTH_MAPPING.items()
            if re.match(pattern=f"^{pattern}$", string=model_config.model_id)
        ]
        hf_model_config.model_max_length = model_lengths[0] if model_lengths else -1

        num_params = [
            num_param
            for pattern, num_param in NUM_PARAMS_MAPPING.items()
            if re.match(pattern=pattern, string=model_config.model_id)
        ]
        hf_model_config.num_params = num_params[0] if num_params else -1

        hf_model_config.id2label = dataset_config.id2label
        hf_model_config.label2id = dataset_config.label2id
        hf_model_config.eos_token_id = hf_model_config.vocab_size - 1
        hf_model_config.bos_token_id = hf_model_config.vocab_size - 1
        hf_model_config.pad_token_id = hf_model_config.vocab_size - 1

        # Check if the vocab size is correct, and if not then correct it
        tok = OpenAITokenizer(
            model_config=model_config, hf_model_config=hf_model_config
        )
        for idx in range(hf_model_config.vocab_size - 1, 0, -1):
            try:
                tok.decode([idx])
                hf_model_config.vocab_size = idx + 1
                break
            except Exception:
                pass
        else:
            raise ValueError(
                f"Couldn't find vocab size for the model {model_config.model_id!r}"
            )

        tokenizer = OpenAITokenizer(
            model_config=model_config, hf_model_config=hf_model_config
        )
        model = OpenAIModel(
            model_config=model_config,
            hf_model_config=hf_model_config,
            dataset_config=dataset_config,
            benchmark_config=self.benchmark_config,
            tokenizer=tokenizer,
        )

        # If the model is a chat model then we need to reduce the maximum context
        # length by 7 tokens, as these are used in the chat prompt
        if model.is_chat_model:
            hf_model_config.model_max_length -= 7
            tokenizer.hf_model_config = hf_model_config
            model.config = hf_model_config

        return model, tokenizer



"""Factory class for creating dataset configurations."""

import importlib.util
import logging
import os
import sys
from typing import TYPE_CHECKING

import torch

from .config import BenchmarkConfig
from .dataset_configs import get_all_dataset_configs
from .enums import Device, Framework
from .exceptions import InvalidBenchmark
from .languages import get_all_languages
from .tasks import get_all_tasks

if TYPE_CHECKING:
    from .config import Language, Task


logger = logging.getLogger(__package__)


def build_benchmark_config(
    progress_bar: bool,
    save_results: bool,
    task: str | list[str] | None,
    dataset: str | list[str] | None,
    language: str | list[str],
    model_language: str | list[str] | None,
    dataset_language: str | list[str] | None,
    framework: Framework | str | None,
    device: Device | None,
    batch_size: int,
    evaluate_train: bool,
    raise_errors: bool,
    cache_dir: str,
    token: bool | str | None,
    openai_api_key: str | None,
    prefer_azure: bool,
    azure_openai_api_key: str | None,
    azure_openai_endpoint: str | None,
    azure_openai_api_version: str | None,
    force: bool,
    verbose: bool,
    trust_remote_code: bool,
    load_in_4bit: bool | None,
    use_flash_attention: bool | None,
    clear_model_cache: bool,
    only_validation_split: bool,
    few_shot: bool,
    num_iterations: int,
    run_with_cli: bool,
    first_time: bool = False,
) -> BenchmarkConfig:
    """Create a benchmark configuration.

    Args:
        progress_bar:
            Whether to show a progress bar when running the benchmark.
        save_results:
            Whether to save the benchmark results to a file.
        task:
            The tasks to include for dataset. If None then datasets will not be
            filtered based on their task.
        dataset:
            The datasets to include for task. If None then all datasets will be
            included, limited by the `task` parameter.
        language:
            The language codes of the languages to include, both for models and
            datasets. Here 'no' means both Bokmål (nb) and Nynorsk (nn). Set this
            to 'all' if all languages (also non-Scandinavian) should be considered.
        model_language:
            The language codes of the languages to include for models. If None then
            the `language` parameter will be used.
        dataset_language:
            The language codes of the languages to include for datasets. If None then
            the `language` parameter will be used.
        framework:
            The framework to use for running the models. If None then the framework
            will be set automatically.
        device:
            The device to use for running the models. If None then the device will be
            set automatically.
        batch_size:
            The batch size to use for running the models.
        evaluate_train:
            Whether to evaluate the models on the training set.
        raise_errors:
            Whether to raise errors when running the benchmark.
        cache_dir:
            The directory to use for caching the models.
        token:
            The token to use for running the models.
        openai_api_key:
            The OpenAI API key to use for running the models.
        prefer_azure:
            Whether to prefer the Azure OpenAI API for running the models, over the
            OpenAI API.
        azure_openai_api_key:
            The Azure OpenAI API key to use for running the models.
        azure_openai_endpoint:
            The Azure OpenAI endpoint to use for running the models.
        azure_openai_api_version:
            The Azure OpenAI api version to use for running the models.
        force:
            Whether to force the benchmark to run even if the results are already
            cached.
        verbose:
            Whether to print verbose output when running the benchmark.
        trust_remote_code:
            Whether to trust remote code when running the benchmark.
        load_in_4bit:
            Whether to load the models in 4-bit precision.
        use_flash_attention:
            Whether to use Flash Attention for the models. If None then it will be used
            if it is available.
        clear_model_cache:
            Whether to clear the model cache before running the benchmark.
        only_validation_split:
            Whether to only use the validation split for the datasets.
        few_shot:
            Whether to use few-shot learning for the models.
        num_iterations:
            The number of iterations each model should be evaluated for.
        run_with_cli:
            Whether the benchmark is being run with the CLI.
        first_time:
            Whether this is the first time the benchmark configuration is being created.
            Defaults to False.

    Returns:
        The benchmark configuration.
    """
    language_codes = get_correct_language_codes(language_codes=language)
    model_languages = prepare_languages(
        language_codes=model_language, default_language_codes=language_codes
    )
    dataset_languages = prepare_languages(
        language_codes=dataset_language, default_language_codes=language_codes
    )

    tasks, datasets = prepare_tasks_and_datasets(
        task=task, dataset=dataset, dataset_languages=dataset_languages
    )

    torch_device = prepare_device(device=device)

    if openai_api_key is None:
        openai_api_key = os.getenv("OPENAI_API_KEY")
    if azure_openai_api_key is None:
        azure_openai_api_key = os.getenv("AZURE_OPENAI_API_KEY")
    if azure_openai_endpoint is None:
        azure_openai_endpoint = os.getenv("AZURE_OPENAI_ENDPOINT")
    if azure_openai_api_version is None:
        azure_openai_api_version = os.getenv("AZURE_OPENAI_API_VERSION")

    # Ensure that we are not using both OpenAI and Azure OpenAI API keys
    if all(
        value is not None
        for value in (
            openai_api_key,
            azure_openai_api_key,
            azure_openai_endpoint,
            azure_openai_api_version,
        )
    ):
        if prefer_azure:
            logger.info(
                "Both OpenAI and Azure OpenAI API keys are set. Using Azure OpenAI."
            )
            openai_api_key = None
        else:
            if run_with_cli:
                logger.info(
                    "Both OpenAI and Azure OpenAI API keys are set. Using OpenAI since "
                    "the `--prefer-azure` flag is not set."
                )
            else:
                logger.info(
                    "Both OpenAI and Azure OpenAI API keys are set. Using OpenAI since "
                    "the `prefer_azure` argument is not set to True."
                )
            azure_openai_api_key = None
            azure_openai_endpoint = None

    # Sanity check
    assert not (openai_api_key is not None and azure_openai_api_key is not None)

    framework_obj = Framework(framework) if framework is not None else None

    if token is True:
        token = None

    if use_flash_attention is None:
        if torch_device.type != "cuda":
            use_flash_attention = False
        elif importlib.util.find_spec("flash_attn") is None:
            use_flash_attention = False
            if first_time and torch_device.type == "cuda":
                message = (
                    "Flash attention has not been installed, so this will not be used. "
                    "To install it, run `pip install -U wheel && "
                    "FLASH_ATTENTION_SKIP_CUDA_BUILD=TRUE pip install flash-attn "
                    "--no-build-isolation`. Alternatively, you can disable this "
                    "message by setting "
                )
                if run_with_cli:
                    message += "the flag `--no-use-flash-attention`."
                else:
                    message += (
                        "the argument `use_flash_attention=False` in the `Benchmarker`."
                    )
                logger.info(message)

    # Set variable with number of iterations
    if hasattr(sys, "_called_from_test"):
        num_iterations = 1

    return BenchmarkConfig(
        model_languages=model_languages,
        dataset_languages=dataset_languages,
        tasks=tasks,
        datasets=datasets,
        batch_size=batch_size,
        raise_errors=raise_errors,
        cache_dir=cache_dir,
        evaluate_train=evaluate_train,
        token=token,
        openai_api_key=openai_api_key,
        azure_openai_api_key=azure_openai_api_key,
        azure_openai_endpoint=azure_openai_endpoint,
        azure_openai_api_version=azure_openai_api_version,
        force=force,
        progress_bar=progress_bar,
        save_results=save_results,
        verbose=verbose,
        framework=framework_obj,
        device=torch_device,
        trust_remote_code=trust_remote_code,
        load_in_4bit=load_in_4bit,
        use_flash_attention=use_flash_attention,
        clear_model_cache=clear_model_cache,
        only_validation_split=only_validation_split,
        few_shot=few_shot,
        num_iterations=num_iterations,
        run_with_cli=run_with_cli,
    )


def get_correct_language_codes(language_codes: str | list[str]) -> list[str]:
    """Get correct language code(s).

    Args:
        language_codes:
            The language codes of the languages to include, both for models and
            datasets. Here 'no' means both Bokmål (nb) and Nynorsk (nn). Set this
            to 'all' if all languages (also non-Scandinavian) should be considered.

    Returns:
        The correct language codes.
    """
    # Create a dictionary that maps languages to their associated language objects
    language_mapping = get_all_languages()

    # Create the list `languages`
    if "all" in language_codes:
        languages = list(language_mapping.keys())
    elif isinstance(language_codes, str):
        languages = [language_codes]
    else:
        languages = language_codes

    # If `languages` contains 'no' then also include 'nb' and 'nn'. Conversely, if
    # either 'nb' or 'nn' are specified then also include 'no'.
    if "no" in languages:
        languages = list(set(languages) | {"nb", "nn"})
    elif "nb" in languages or "nn" in languages:
        languages = list(set(languages) | {"no"})

    return languages


def prepare_languages(
    language_codes: str | list[str] | None, default_language_codes: list[str]
) -> list["Language"]:
    """Prepare language(s) for benchmarking.

    Args:
        language_codes:
            The language codes of the languages to include for models or datasets.
            If specified then this overrides the `language` parameter for model or
            dataset languages.
        default_language_codes:
            The default language codes of the languages to include.

    Returns:
        The prepared model or dataset languages.
    """
    # Create a dictionary that maps languages to their associated language objects
    language_mapping = get_all_languages()

    # Create the list `languages_str` of language codes to use for models or datasets
    languages_str: list[str]
    if language_codes is None:
        languages_str = default_language_codes
    elif isinstance(language_codes, str):
        languages_str = [language_codes]
    else:
        languages_str = language_codes

    # Convert the model languages to language objects
    if "all" in languages_str:
        prepared_languages = list(language_mapping.values())
    else:
        prepared_languages = [language_mapping[language] for language in languages_str]

    return prepared_languages


def prepare_tasks_and_datasets(
    task: str | list[str] | None,
    dataset_languages: list["Language"],
    dataset: str | list[str] | None,
) -> tuple[list["Task"], list[str]]:
    """Prepare task(s) and dataset(s) for benchmarking.

    Args:
        task:
            The tasks to include for dataset. If None then datasets will not be
            filtered based on their task.
        dataset_languages:
            The languages of the datasets in the benchmark.
        dataset:
            The datasets to include for task. If None then all datasets will be
            included, limited by the `task` and `dataset_languages` parameters.

    Returns:
        The prepared tasks and datasets.

    Raises:
        InvalidBenchmark:
            If the task or dataset is not found in the benchmark tasks or datasets.
    """
    # Create a dictionary that maps benchmark tasks to their associated benchmark
    # task objects, and a dictionary that maps dataset names to their associated
    # dataset configuration objects
    task_mapping = get_all_tasks()
    all_dataset_configs = get_all_dataset_configs()

    # Create the list of dataset tasks
    try:
        if task is None:
            tasks = list(task_mapping.values())
        elif isinstance(task, str):
            tasks = [task_mapping[task]]
        else:
            tasks = [task_mapping[t] for t in task]
    except KeyError as e:
        raise InvalidBenchmark(f"Task {e} not found in the benchmark tasks.") from e

    all_official_datasets = [
        dataset_name
        for dataset_name, dataset_config in all_dataset_configs.items()
        if not dataset_config.unofficial
    ]
    if dataset is None:
        dataset = all_official_datasets
    elif isinstance(dataset, str):
        dataset = [dataset]

    all_datasets = list(all_dataset_configs.keys())
    invalid_datasets = set(dataset) - set(all_datasets)
    if invalid_datasets:
        raise InvalidBenchmark(
            f"Dataset(s) {', '.join(invalid_datasets)} not found in the benchmark "
            "datasets."
        )

    datasets = [
        dataset_name
        for dataset_name, dataset_config in all_dataset_configs.items()
        if dataset_name in dataset
        and dataset_config.task in tasks
        and set(dataset_config.languages).intersection(dataset_languages)
    ]

    return tasks, datasets


def prepare_device(device: Device | None) -> torch.device:
    """Prepare device for benchmarking.

    Args:
        device:
            The device to use for running the models. If None then the device will be
            set automatically.

    Returns:
        The prepared device.
    """
    device_mapping = {
        Device.CPU: torch.device("cpu"),
        Device.CUDA: torch.device("cuda"),
        Device.MPS: torch.device("mps"),
    }
    if isinstance(device, Device):
        return device_mapping[device]

    if torch.cuda.is_available():
        return torch.device("cuda")
    elif torch.backends.mps.is_available():
        return torch.device("mps")
    else:
        return torch.device("cpu")



"""Enums used in the project."""

from enum import Enum, auto


class AutoStrEnum(str, Enum):
    """StrEnum where auto() returns the field name in lower case."""

    @staticmethod
    def _generate_next_value_(
        name: str, start: int, count: int, last_values: list
    ) -> str:
        return name.lower()


class Device(AutoStrEnum):
    """The compute device to use for the evaluation.

    Attributes:
        CPU:
            CPU device.
        MPS:
            MPS GPU, used in M-series MacBooks.
        CUDA:
            CUDA GPU, used with NVIDIA GPUs.
    """

    CPU = auto()
    MPS = auto()
    CUDA = auto()


class Framework(AutoStrEnum):
    """The framework of a model.

    Attributes:
        PYTORCH:
            PyTorch framework.
        JAX:
            JAX framework.
        API:
            Accessible via an API.
        HUMAN:
            Human evaluator.
    """

    PYTORCH = auto()
    JAX = auto()
    API = auto()
    HUMAN = auto()


class ModelType(AutoStrEnum):
    """The type of a model.

    Attributes:
        FRESH:
            Randomly initialised Hugging Face model.
        HF:
            Model from the Hugging Face Hub.
        LOCAL:
            Locally stored Hugging Face model.
        OPENAI:
            Model from OpenAI.
        HUMAN:
            Human evaluator.
    """

    FRESH = auto()
    HF = auto()
    LOCAL = auto()
    OPENAI = auto()
    HUMAN = auto()


class DataType(AutoStrEnum):
    """The data type of the model weights.

    Attributes:
        FP32:
            32-bit floating point.
        FP16:
            16-bit floating point.
        BF16:
            16-bit bfloat.
    """

    FP32 = auto()
    FP16 = auto()
    BF16 = auto()



"""List of languages and their ISO 639-1 codes.

Taken from https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes.

Last updated 19 June 2022.
"""

from .config import Language


def get_all_languages() -> dict[str, Language]:
    """Get a list of all the languages.

    Returns:
        A mapping between language codes and their configurations.
    """
    return {cfg.code: cfg for cfg in globals().values() if isinstance(cfg, Language)}


AB = Language(code="ab", name="Abkhazian")
AA = Language(code="aa", name="Afar")
AF = Language(code="af", name="Afrikaans")
SQ = Language(code="sq", name="Albanian")
AM = Language(code="am", name="Amharic")
AR = Language(code="ar", name="Arabic")
AN = Language(code="an", name="Aragonese")
HY = Language(code="hy", name="Armenian")
AS = Language(code="as", name="Assamese")
AV = Language(code="av", name="Avaric")
AE = Language(code="ae", name="Avestan")
AY = Language(code="ay", name="Aymara")
AZ = Language(code="az", name="Azerbaijani")
BM = Language(code="bm", name="Bambara")
BA = Language(code="ba", name="Bashkir")
EU = Language(code="eu", name="Basque")
BE = Language(code="be", name="Belarusian")
BN = Language(code="bn", name="Bengali")
BI = Language(code="bi", name="Bislama")
BS = Language(code="bs", name="Bosnian")
BR = Language(code="br", name="Breton")
BG = Language(code="bg", name="Bulgarian")
MY = Language(code="my", name="Burmese")
CA = Language(code="ca", name="Catalan")
CH = Language(code="ch", name="Chamorro")
CE = Language(code="ce", name="Chechen")
NY = Language(code="ny", name="Chichewa")
ZH = Language(code="zh", name="Chinese")
CU = Language(code="cu", name="Church Slavic")
CV = Language(code="cv", name="Chuvash")
KW = Language(code="kw", name="Cornish")
CO = Language(code="co", name="Corsican")
CR = Language(code="cr", name="Cree")
HR = Language(code="hr", name="Croatian")
CS = Language(code="cs", name="Czech")
DA = Language(code="da", name="Danish")
DV = Language(code="dv", name="Divehi")
NL = Language(code="nl", name="Dutch")
DZ = Language(code="dz", name="Dzongkha")
EN = Language(code="en", name="English")
EO = Language(code="eo", name="Esperanto")
ET = Language(code="et", name="Estonian")
EE = Language(code="ee", name="Ewe")
FO = Language(code="fo", name="Faroese")
FJ = Language(code="fj", name="Fijian")
FI = Language(code="fi", name="Finnish")
FR = Language(code="fr", name="French")
FY = Language(code="fy", name="Western Frisian")
FF = Language(code="ff", name="Fulah")
GD = Language(code="gd", name="Gaelic")
GL = Language(code="gl", name="Galician")
LG = Language(code="lg", name="Ganda")
KA = Language(code="ka", name="Georgian")
DE = Language(code="de", name="German")
EL = Language(code="el", name="Greek")
KL = Language(code="kl", name="Greenlandic")
GN = Language(code="gn", name="Guarani")
GU = Language(code="gu", name="Gujarati")
HT = Language(code="ht", name="Haitian")
HA = Language(code="ha", name="Hausa")
HE = Language(code="he", name="Hebrew")
HZ = Language(code="hz", name="Herero")
HI = Language(code="hi", name="Hindi")
HO = Language(code="ho", name="Hiri Motu")
HU = Language(code="hu", name="Hungarian")
IS = Language(code="is", name="Icelandic")
IO = Language(code="io", name="Ido")
IG = Language(code="ig", name="Igbo")
ID = Language(code="id", name="Indonesian")
IA = Language(code="ia", name="Interlingua")
IE = Language(code="ie", name="Interlingue")
IU = Language(code="iu", name="Inuktitut")
IK = Language(code="ik", name="Inupiaq")
GA = Language(code="ga", name="Irish")
IT = Language(code="it", name="Italian")
JA = Language(code="ja", name="Japanese")
KN = Language(code="kn", name="Kannada")
KR = Language(code="kr", name="Kanuri")
KS = Language(code="ks", name="Kashmiri")
KK = Language(code="kk", name="Kazakh")
KM = Language(code="km", name="Central Khmer")
KI = Language(code="ki", name="Kikuyu")
RW = Language(code="rw", name="Kinyarwanda")
KY = Language(code="ky", name="Kirghiz")
KV = Language(code="kv", name="Komi")
KG = Language(code="kg", name="Kongo")
KO = Language(code="ko", name="Korean")
KJ = Language(code="kj", name="Kuanyama")
KU = Language(code="ku", name="Kurdish")
LO = Language(code="lo", name="Lao")
LA = Language(code="la", name="Latin")
LV = Language(code="lv", name="Latvian")
LI = Language(code="li", name="Limburgan")
LN = Language(code="ln", name="Lingala")
LT = Language(code="lt", name="Lithuanian")
LU = Language(code="lu", name="Luba-Katanga")
LB = Language(code="lb", name="Luxembourgish")
MK = Language(code="mk", name="Macedonian")
MG = Language(code="mg", name="Malagasy")
MS = Language(code="ms", name="Malay")
ML = Language(code="ml", name="Malayalam")
MT = Language(code="mt", name="Maltese")
GV = Language(code="gv", name="Manx")
MI = Language(code="mi", name="Maori")
MR = Language(code="mr", name="Marathi")
MH = Language(code="mh", name="Marshallese")
MN = Language(code="mn", name="Mongolian")
NA = Language(code="na", name="Nauru")
NV = Language(code="nv", name="Navajo")
ND = Language(code="nd", name="Northern Ndebele")
NR = Language(code="nr", name="South Ndebele")
NG = Language(code="ng", name="Ndonga")
NE = Language(code="ne", name="Nepali")
NO = Language(code="no", name="Norwegian")
NB = Language(code="nb", name="Norwegian Bokmål")
NN = Language(code="nn", name="Norwegian Nynorsk")
II = Language(code="ii", name="Sichuan Yi")
OC = Language(code="oc", name="Occitan")
OJ = Language(code="oj", name="Ojibwa")
OR = Language(code="or", name="Oriya")
OM = Language(code="om", name="Oromo")
OS = Language(code="os", name="Ossetian")
PI = Language(code="pi", name="Pali")
PS = Language(code="ps", name="Pashto")
FA = Language(code="fa", name="Persian")
PL = Language(code="pl", name="Polish")
PT = Language(code="pt", name="Portuguese")
PA = Language(code="pa", name="Punjabi")
QU = Language(code="qu", name="Quechua")
RO = Language(code="ro", name="Romanian")
RM = Language(code="rm", name="Romansh")
RN = Language(code="rn", name="Rundi")
RU = Language(code="ru", name="Russian")
SE = Language(code="se", name="Northern Sami")
SM = Language(code="sm", name="Samoan")
SG = Language(code="sg", name="Sango")
SA = Language(code="sa", name="Sanskrit")
SC = Language(code="sc", name="Sardinian")
SR = Language(code="sr", name="Serbian")
SN = Language(code="sn", name="Shona")
SD = Language(code="sd", name="Sindhi")
SI = Language(code="si", name="Sinhala")
SK = Language(code="sk", name="Slovak")
SL = Language(code="sl", name="Slovenian")
SO = Language(code="so", name="Somali")
ST = Language(code="st", name="Sotho")
ES = Language(code="es", name="Spanish")
SU = Language(code="su", name="Sundanese")
SW = Language(code="sw", name="Swahili")
SS = Language(code="ss", name="Swati")
SV = Language(code="sv", name="Swedish")
TL = Language(code="tl", name="Tagalog")
TY = Language(code="ty", name="Tahitian")
TG = Language(code="tg", name="Tajik")
TA = Language(code="ta", name="Tamil")
TT = Language(code="tt", name="Tatar")
TE = Language(code="te", name="Telugu")
TH = Language(code="th", name="Thai")
BO = Language(code="bo", name="Tibetan")
TI = Language(code="ti", name="Tigrinya")
TO = Language(code="to", name="Tonga")
TS = Language(code="ts", name="Tsonga")
TN = Language(code="tn", name="Tswana")
TR = Language(code="tr", name="Turkish")
TK = Language(code="tk", name="Turkmen")
TW = Language(code="tw", name="Twi")
UG = Language(code="ug", name="Uighur")
UK = Language(code="uk", name="Ukrainian")
UR = Language(code="ur", name="Urdu")
UZ = Language(code="uz", name="Uzbek")
VE = Language(code="ve", name="Venda")
VI = Language(code="vi", name="Vietnamese")
VO = Language(code="vo", name="Volapük")
WA = Language(code="wa", name="Walloon")
CY = Language(code="cy", name="Welsh")
WO = Language(code="wo", name="Wolof")
XH = Language(code="xh", name="Xhosa")
YI = Language(code="yi", name="Yiddish")
YO = Language(code="yo", name="Yoruba")
ZA = Language(code="za", name="Zhuang")
ZU = Language(code="zu", name="Zulu")



"""Functions related to the loading of models."""

from typing import TYPE_CHECKING, Type

from .exceptions import InvalidBenchmark
from .model_setups import (
    FreshModelSetup,
    HFModelSetup,
    LocalModelSetup,
    OpenAIModelSetup,
)
from .utils import (
    GENERATIVE_DATASET_SUPERTASKS,
    GENERATIVE_DATASET_TASKS,
    model_is_generative,
)

if TYPE_CHECKING:
    from transformers import PreTrainedModel

    from .config import BenchmarkConfig, DatasetConfig, ModelConfig
    from .protocols import GenerativeModel, ModelSetup, Tokenizer


def load_model(
    model_config: "ModelConfig",
    dataset_config: "DatasetConfig",
    benchmark_config: "BenchmarkConfig",
) -> tuple["PreTrainedModel | GenerativeModel", "Tokenizer"]:
    """Load a model.

    Args:
        model_config:
            The model configuration.
        dataset_config:
            The dataset configuration.
        benchmark_config:
            The benchmark configuration.

    Returns:
        The tokenizer and model.
    """
    model_type_to_model_setup_mapping: dict[str, Type["ModelSetup"]] = dict(
        fresh=FreshModelSetup,
        hf=HFModelSetup,
        local=LocalModelSetup,
        openai=OpenAIModelSetup,
    )
    setup_class = model_type_to_model_setup_mapping[model_config.model_type]
    setup = setup_class(benchmark_config=benchmark_config)

    model, tokenizer = setup.load_model(
        model_config=model_config, dataset_config=dataset_config
    )

    # Refuse to benchmark non-generative models on generative tasks
    if (
        (
            dataset_config.task.supertask in GENERATIVE_DATASET_SUPERTASKS
            or dataset_config.task.name in GENERATIVE_DATASET_TASKS
        )
        and model is not None
        and not model_is_generative(model=model)
    ):
        raise InvalidBenchmark(
            f"Cannot benchmark non-generative model {model_config.model_id!r} on "
            f"generative task {dataset_config.task.name!r}."
        )

    # TODO: XMOD model setup: https://huggingface.co/facebook/xmod-base#input-language

    return model, tokenizer
